/* $Id: mga_tritemp.h,v 1.1 1999/07/31 07:07:55 crow Exp $ */
/*
 * This file has been modified by Wittawat Yamwong for GLX module.
 */
/*
 * Mesa 3-D graphics library
 * Version:  3.1
 * 
 * Copyright (C) 1999  Brian Paul   All Rights Reserved.
 * 
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * BRIAN PAUL BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN
 * AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/*
 * Triangle Rasterizer Template
 *
 * This file is #include'd to generate custom triangle rasterizers.
 *
 * The following macros may be defined to indicate what auxillary information
 * must be interplated across the triangle:
 *    INTERP_Z      - if defined, interpolate Z values
 *    INTERP_RGB    - if defined, interpolate RGB values
 *    INTERP_SPEC   - if defined, interpolate specular RGB values
 *    INTERP_ALPHA  - if defined, interpolate Alpha values
 *    INTERP_INDEX  - if defined, interpolate color index values
 *    INTERP_STUV   - if defined, interpolate set 0 float STRQ texcoords
 *                         NOTE:  OpenGL STRQ = Mesa STUV (R was taken for red)
 *
 * Optionally, one may provide one-time setup code per triangle:
 *    SETUP_CODE    - code which is to be executed once per triangle
 * 
 * This code was designed for the origin to be in the lower-left corner.
 *
 * Inspired by triangle rasterizer code written by Allen Akin.  Thanks Allen!
 */
 
/*
 * Use "flipped y" coordinate system. (0,0) at the top left corner. 
 */


/*
void triangle( GLcontext *ctx, GLuint v0, GLuint v1, GLuint v2, GLuint pv )
{
*/
typedef struct {
  mgaVertexPtr v0, v1; /* Y(v0) < Y(v1) */
  GLfloat dx;          /* X(v1) - X(v0) */
  GLfloat dy;          /* Y(v1) - Y(v0) */
  GLfloat dxOOA;          /* dx * oneOverArea */
  GLfloat dyOOA;          /* dy * oneOverArea */
  
  GLfloat adjx,adjy; /* subpixel offset after rounding to integer */
  mgaI32 err;        /* error term ready for hardware */
  mgaI32 idx,idy;    /* delta-x & delta-y ready for hardware */
  mgaI32 sx,sy;	     /* first sample point x,y coord */
  mgaI32 lines;      /* number of lines to be sampled on this edge */
} EdgeT;

#define F2COL(x) (RINT(x) & 0x00ffffff)

#ifndef INTERP_RGB
struct vertex_buffer *mesaVB = ctx->VB;
#endif
mgaVertexPtr	VB = mgaCtx->VB;
EdgeT		eMaj, eTop, eBot;
GLfloat		oneOverArea;
mgaVertexPtr	vMin, vMid, vMax; /* Y(vMin)<=Y(vMid)<=Y(vMax) */
int 		Shape; /* 1 = Top half, 2 = bottom half, 3 = top+bottom */
DMALOCALS;

	/* bump the performance counter */
	mgaglx.c_triangles++;

	/* setup 3D state if needed */
	ENTER3D;

/* find the order of the 3 vertices along the Y axis */
{
  GLfloat y0 = VB[v0].y;
  GLfloat y1 = VB[v1].y;
  GLfloat y2 = VB[v2].y;
 
  if (y0<=y1) {
    if (y1<=y2) {
      vMin = &VB[v0];   vMid = &VB[v1];   vMax = &VB[v2];   /* y0<=y1<=y2 */
    }
    else if (y2<=y0) {
      vMin = &VB[v2];   vMid = &VB[v0];   vMax = &VB[v1];   /* y2<=y0<=y1 */
    }
    else {
      vMin = &VB[v0];   vMid = &VB[v2];   vMax = &VB[v1];   /* y0<=y2<=y1 */
    }
  }
  else {
    if (y0<=y2) {
      vMin = &VB[v1];   vMid = &VB[v0];   vMax = &VB[v2];   /* y1<=y0<=y2 */
    }
    else if (y2<=y1) {
      vMin = &VB[v2];   vMid = &VB[v1];   vMax = &VB[v0];   /* y2<=y1<=y0 */
    }
    else {
      vMin = &VB[v1];   vMid = &VB[v2];   vMax = &VB[v0];   /* y1<=y2<=y0 */
    }
  }
}

/* vertex/edge relationship */
eMaj.v0 = vMin;   eMaj.v1 = vMax;
eTop.v0 = vMin;   eTop.v1 = vMid;
eBot.v0 = vMid;   eBot.v1 = vMax;

/* compute deltas for each edge:  vertex[v1] - vertex[v0] */
eMaj.dx = vMax->x - vMin->x;
eMaj.dy = vMax->y - vMin->y;
eTop.dx = vMid->x - vMin->x;
eTop.dy = vMid->y - vMin->y;
eBot.dx = vMax->x - vMid->x;
eBot.dy = vMax->y - vMid->y;
 
/* compute oneOverArea */
{
  GLfloat area = eMaj.dx * eBot.dy - eBot.dx * eMaj.dy;
  if (area>-0.05f && area<0.05f) {
    return;  /* very small; CULLED */
  }
  oneOverArea = 1.0F / area;
}

/* Edge setup.  For a triangle strip these could be reused... */
{

#define DELTASCALE 16 /* Scaling factor for idx and idy. Note that idx and
                         idy are 18 bits signed, so don't choose too big
                         value. */
    
  mgaI32  ivMax_y;
  GLfloat temp;

  ivMax_y = vMax->ceilY;
  eTop.sy = eMaj.sy = vMin->ceilY;
  eBot.sy = vMid->ceilY;
  
  eMaj.lines = ivMax_y - eMaj.sy;
  if (eMaj.lines > 0) {
    GLfloat dxdy = eMaj.dx / eMaj.dy;
    eMaj.adjy = (GLfloat) eMaj.sy - vMin->y;
    temp = vMin->x + eMaj.adjy*dxdy;
    eMaj.sx = CEIL(temp);
    eMaj.adjx = (GLfloat) eMaj.sx - vMin->x;
    if (eMaj.lines == 1) {
      eMaj.idy = 1;
      eMaj.idx = 0;
      eMaj.err = 0;
    } else {
      eMaj.idy = RINT(eMaj.dy * DELTASCALE);
      eMaj.idx = FLOOR(eMaj.idy * dxdy);
      eMaj.err = RINT(((GLfloat) eMaj.sx - temp) * (GLfloat)eMaj.idy);
    }
  } else {
    return; /* CULLED */
  }

  Shape = 3;

  eBot.lines = ivMax_y - eBot.sy;
  if (eBot.lines > 0) {
    GLfloat dxdy = eBot.dx / eBot.dy;
    eBot.adjy = (GLfloat) eBot.sy - vMid->y;
    temp = vMid->x + eBot.adjy*dxdy;
    eBot.sx = CEIL(temp);
    eBot.adjx = (GLfloat) eBot.sx - vMid->x;
    if (eBot.lines == 1) {
      eBot.idy = 1;
      eBot.idx = 0;
      eBot.err = 0;
    } else {
      eBot.idy = RINT(eBot.dy * DELTASCALE);
      eBot.idx = FLOOR(eBot.idy * dxdy);
      eBot.err = RINT(((GLfloat) eBot.sx - temp) * (GLfloat)eBot.idy);
    }
  } else {
    Shape = 1;
  }

  eTop.lines = eBot.sy - eTop.sy;
  if (eTop.lines > 0) {
    GLfloat dxdy = eTop.dx / eTop.dy;
    eTop.adjy = eMaj.adjy;
    temp = vMin->x + eTop.adjy*dxdy;
    eTop.sx = CEIL(temp);
    eTop.adjx = (GLfloat) eTop.sx - vMin->x;
    if (eTop.lines == 1) {
      eTop.idy = 1;
      if (eBot.lines > 0) {
        eTop.idx = eBot.sx - eTop.sx; /* needed for bottom half */
      } else {
        eTop.idx = 0;
      }
      eTop.err = 0;
    } else {
      eTop.idy = RINT(eTop.dy * DELTASCALE);
      eTop.idx = FLOOR(eTop.idy * dxdy);
      eTop.err = RINT(((GLfloat) eTop.sx - temp) * (GLfloat)eTop.idy);
    }
  } else {
    Shape = 2;
  }
}

{
  GLint ltor;		 /* true if scanning left-to-right */
  EdgeT *eLeft, *eRight;
  int lines;
  mgaVertexPtr vTL;      /* Top left vertex */
  GLfloat adjx, adjy;

  /*
   * Execute user-supplied setup code
   */
#ifdef SETUP_CODE
  SETUP_CODE
#endif

  ltor = (oneOverArea > 0.0F);
  
  if (Shape == 2) {
    /* bottom half triangle */
    if (ltor) {
      eLeft = &eMaj;
      eRight = &eBot;
    } else {
      eLeft = &eBot;
      eRight = &eMaj;
    }
    lines = eBot.lines;
  } else {
    /* top half triangle */
    if (ltor) {
      eLeft = &eMaj;
      eRight = &eTop;
    } else {
      eLeft = &eTop;
      eRight = &eMaj;
    }
    lines = eTop.lines;
  }
    
  vTL = eLeft->v0;
  adjx = eLeft->adjx; adjy = eLeft->adjy;

/* prepare to write dma commands */
MGADMAGETPTR(75);


  /* setup derivatives */
/* compute d?/dx and d?/dy derivatives */
eBot.dxOOA = eBot.dx * oneOverArea;
eBot.dyOOA = eBot.dy * oneOverArea;
eMaj.dxOOA = eMaj.dx * oneOverArea;
eMaj.dyOOA = eMaj.dy * oneOverArea;

#define DERIV( DZ, COMP) \
  { \
    GLfloat eMaj_DZ, eBot_DZ; \
    eMaj_DZ = vMax->COMP - vMin->COMP; \
    eBot_DZ = vMax->COMP - vMid->COMP; \
    DZ ## dx = eMaj_DZ * eBot.dyOOA - eMaj.dyOOA * eBot_DZ;    \
    DZ ## dy = eMaj.dxOOA * eBot_DZ - eMaj_DZ * eBot.dxOOA;    \
  }

#if INTERP_Z
  if (!mgaCtx->NoZ) {
    GLfloat Zstart;
    GLfloat dzdx, dzdy;

    DERIV(dz, z);
    if (dzdx>DEPTH_SCALE*(1<<15) || dzdx<-DEPTH_SCALE*(1<<15)) {
      /* probably a sliver triangle */
      dzdx = 0.0;
      dzdy = 0.0;
    }
    Zstart = vTL->z + dzdx*adjx + dzdy*adjy + 
             ctx->PolygonZoffset * (1 << 15);
#if DEPTH_BITS==16
    if (Zstart > 65535.0F*(1 << 15)) {
      Zstart = 65535.0F*(1 << 15);
      dzdx = 0.0F;
      dzdy = 0.0F; 
    }
    MGADMA_Z16(RINT(Zstart),RINT(dzdx),RINT(dzdy));
#else
#error 32 bits depth buffer not yet supported
#endif
  } /* endif (!mgaCtx->NoZ) */
#endif /* INTERP_Z */

#if INTERP_RGB
  {
    GLfloat drdx, drdy;
    GLfloat dgdx, dgdy;
    GLfloat dbdx, dbdy;

    DERIV(dr, r);
      MGADMA_R(F2COL(vTL->r+drdx*adjx+drdy*adjy),F2COL(drdx),F2COL(drdy));
    DERIV(dg, g);
      MGADMA_G(F2COL(vTL->g+dgdx*adjx+dgdy*adjy),F2COL(dgdx),F2COL(dgdy));
    DERIV(db, b);
      MGADMA_B(F2COL(vTL->b+dbdx*adjx+dbdy*adjy),F2COL(dbdx),F2COL(dbdy));
  }
#else
  {
    GLubyte *col = &(mesaVB->VBCOLOR[pv][0]);
    MGADMA_R(col[0] << 15,0,0);
    MGADMA_G(col[1] << 15,0,0);
    MGADMA_B(col[2] << 15,0,0);
  }
#endif


#if INTERP_ALPHA
  {
    GLfloat dadx, dady;

    DERIV(da, a);
      MGADMA_A(F2COL(vTL->a+dadx*adjx+dady*adjy),F2COL(dadx),F2COL(dady));
      MGADMA_ALPHACTRL(mgaCtx->regALPHACTRL);
  }
#else
  MGADMA_A(mesaVB->VBCOLOR[pv][3] << 15,0,0);
  MGADMA_ALPHACTRL(mgaCtx->regALPHACTRL);
#endif

#if INTERP_SPEC
  {
    GLfloat dsrdx, dsrdy;
    GLfloat dsgdx, dsgdy;
    GLfloat dsbdx, dsbdy;

    DERIV(dsr, sr);	
      MGADMA_SPECR(F2COL(vTL->sr+dsrdx*adjx+dsrdy*adjy),F2COL(dsrdx),F2COL(dsrdy));
    DERIV(dsg, sg);
      MGADMA_SPECG(F2COL(vTL->sg+dsgdx*adjx+dsgdy*adjy),F2COL(dsgdx),F2COL(dsgdy));
    DERIV(dsb, sb);
      MGADMA_SPECB(F2COL(vTL->sb+dsbdx*adjx+dsbdy*adjy),F2COL(dsbdx),F2COL(dsbdy));
  }
#elif FLAT_SPEC
  {
    GLubyte *spec = &(mesaVB->Specular[pv][0]);
    MGADMA_SPECR(spec[0] << 15,0,0);
    MGADMA_SPECG(spec[1] << 15,0,0);
    MGADMA_SPECB(spec[2] << 15,0,0);
  }
#endif


#if INTERP_STUV
  {
    GLfloat dsdx, dsdy;
    GLfloat dtdx, dtdy;
    GLfloat dvdx, dvdy;
  	   
    DERIV(ds,s);
      MGADMA_S(RINT(vTL->s+dsdx*adjx+dsdy*adjy),RINT(dsdx),RINT(dsdy));
    DERIV(dt,t);
      MGADMA_T(RINT(vTL->t+dtdx*adjx+dtdy*adjy),RINT(dtdx),RINT(dtdy));
    DERIV(dv,v); /* can add "if (needed)" for mesa 3.1 */
    {
      mgaI32 sq = RINT(vTL->v+dvdx*adjx+dvdy*adjy);
      MGADMA_Q((sq == 0) ? 1 : sq,RINT(dvdx),RINT(dvdy));
    }
  }
#endif
	
  if (mgaCtx->Fog.Apply) {
    /* approximate fog factor using piecewise linearization */
    GLfloat dfdx,dfdy;
    DERIV(df,fog);
    MGADMA_FOG(F2COL(vTL->fog + dfdx*adjx + dfdy*adjy),
            F2COL(dfdx),F2COL(dfdy),mgaCtx->Fog.Color);
  }
    	
  {
    mgaUI32 sgn = 0;

    /* General setup for the first part */ 
    if (MGA_IS_G400(MGAchipset))
      {
	MGADMA_G400_LSLOPE(eLeft->idx,eLeft->idy,sgn,eLeft->err);
	MGADMA_G400_RSLOPE(eRight->idx,eRight->idy,sgn,eRight->err);
      }
    else
      {
	MGADMA_LSLOPE(eLeft->idx,eLeft->idy,sgn,eLeft->err);
	MGADMA_RSLOPE(eRight->idx,eRight->idy,sgn,eRight->err);	
      }

    MGADMA_SGN(sgn);
    MGADMA_FXBNDRY(eLeft->sx,eRight->sx);
    MGADMA_YDSTLEN(eLeft->sy,lines);

#ifdef DRAW_CODE
    DRAW_CODE;
#endif

    /* Draw part #1 */
    {     
      mgaUI32 dwgctl = mgaCtx->regDWGCTL;
#if TEXTURE
      dwgctl |= DC_shftzero_enable | DC_opcod_texture_trap;
#else
      dwgctl |= DC_shftzero_enable | DC_opcod_trap;
#endif
#if INTERP_Z
      dwgctl |= mgaCtx->zmode;
#else
      dwgctl |= DC_atype_i | DC_zmode_nozcmp;
#endif
      MGADMA_DWGCTL_EXEC(dwgctl);
    }
    
    if (Shape != 3) { /* has only one half? */
      DMAADVANCE();
      return;
    }

    /* Draw part #2 */
    if (ltor) {
      if (MGA_IS_G400(MGAchipset))
	MGADMA_G400_RSLOPE(eBot.idx,eBot.idy,sgn,eBot.err);
      else
	MGADMA_RSLOPE(eBot.idx,eBot.idy,sgn,eBot.err);
	
      DMAOUTREG(MGAREG_FXRIGHT,eBot.sx);
    } else {
#if 0
      GLfloat temp = (eTop.sx-eBot.sx) + 
        ((GLfloat)eTop.idx * (GLfloat)eTop.lines / (GLfloat)eTop.idy - 
         (GLfloat)eTop.err/(GLfloat)eTop.idy);
      if (temp < -1.0) {
        ;
      } else if (temp > 0.0) {
        mgaError("BRKLEFT error, temp=%f",temp);
        mgaError("\t(%f,%f) (%f,%f) (%f,%f)\n",vMax->x,vMax->y,
                 vMid->x,vMid->y,vMin->x,vMin->y);
        mgaError("\tTop sx=%ld, idx=%ld, idy=%ld, err=%ld, sy=%d\n"
                 "\tBot sx=%ld, sy=%d\n",
                 eTop.sx,eTop.idx,eTop.idy,eTop.err,eTop.sy,
                 eBot.sx,eBot.sy);
      }
#endif
      sgn |= S_brkleft_enable;
      DMAOUTREG(MGAREG_FXLEFT,eBot.sx);
      if (MGA_IS_G400(MGAchipset))
	MGADMA_G400_LSLOPE(eBot.idx,eBot.idy,sgn,eBot.err);
      else
	MGADMA_LSLOPE(eBot.idx,eBot.idy,sgn,eBot.err);
	
    }
    MGADMA_SGN(sgn);
    MGADMA_LEN_EXEC(eBot.lines);
    DMAADVANCE();
  }
}
/*}*/

#undef SETUP_CODE

#undef INTERP_Z
#undef INTERP_RGB
#undef INTERP_SPEC
#undef INTERP_ALPHA
#undef INTERP_STUV

#undef TEXTURE
#undef DRAW_CODE
#undef DERIV
#undef DERIV_W
#undef FLAT_SPEC
#undef F2COL
#undef DELTASCALE
