/*
 * GLX Hardware Device Driver for Matrox Millenium G200
 * Copyright (C) 1999 Wittawat Yamwong
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * WITTAWAT YAMWONG, OR ANY OTHER CONTRIBUTORS BE LIABLE FOR ANY CLAIM, 
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE 
 * OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 *
 *    Wittawat Yamwong <Wittawat.Yamwong@stud.uni-hannover.de>
 */

/* $Id: mgacontext.c,v 1.2 1999/08/01 04:07:18 johnc Exp $ */

#include "xsmesaP.h"
#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <GL/gl.h>

#include "mgacontext.h"
#include "mgalib.h"
#include "g200_mac.h"
#include "mgalog.h"

#if defined(MESA31) 
#define DEFAULT2D DefaultD[2]
#else
#define DEFAULT2D Default2D
#endif

mgaContextPtr mgaCreateContext(GLcontext *ctx)
{
  mgaContextPtr c;

  c = (mgaContextPtr) calloc(1,sizeof(mgaContext));
  if (!c)
    return NULL;

  c->magic = mgaContextMagic;
  c->gl_ctx = ctx;
  c->regDWGCTL = DC_clipdis_enable | (0xC << DC_bop_SHIFT);
  c->regALPHACTRL = AC_src_one | AC_dst_zero | AC_amode_FCOL | 
    AC_astipple_disable | AC_aten_disable | AC_atmode_noacmp |
    AC_alphasel_fromtex;
  c->zmode = DC_zmode_nozcmp | DC_atype_zi;
  c->InvWScale = 1.0;

#ifdef MESA31
  c->TextureMode = ctx->Texture.Unit[0].EnvMode;
  c->DefaultTexObj = mgaCreateTexObj(c,ctx->Shared->DEFAULT2D[ctx->Texture.CurrentUnit]);
  c->CurrentTexObj = c->DefaultTexObj;
  ctx->Shared->DEFAULT2D[ctx->Texture.CurrentUnit]->DriverData = c->DefaultTexObj;
#else
  c->TextureMode = ctx->Texture.Set[0].EnvMode;
  c->DefaultTexObj = mgaCreateTexObj(c,ctx->Shared->DEFAULT2D[ctx->Texture.CurrentSet]);
  c->CurrentTexObj = c->DefaultTexObj;
  ctx->Shared->DEFAULT2D[ctx->Texture.CurrentSet]->DriverData = c->DefaultTexObj;
#endif

  mgaInitFogAttrib(&c->Fog);
  if (c->DefaultTexObj == NULL) {
    mgaError("mgaCreateContext(): cannot create DefaultTexObj\n");
    free(c);
    return NULL;
  }

  mgaMsg(2,"mgaCreateContext(): successful.\n");
  return c;
}

int mgaDestroyContext(mgaContextPtr ctx)
{
  if (!ctx)
    return 0;
  if (ctx->magic != mgaContextMagic)
    return -1;
  if (--(ctx->refcount) > 0)
    return 0;
  while (ctx->TexObjList)
    mgaDestroyTexObj(ctx,ctx->TexObjList);
    
    if ( ctx == mgaCtx ) {
    	mgaCtx = NULL;
    }
  ctx->magic = 0;
  free(ctx);
  mgaMsg(2,"mgaDestroyContext(): successfully destroyed.\n");
  return 0;
}
