/*
 * GLX Hardware Device Driver for Matrox Millenium G200
 * Copyright (C) 1999 Wittawat Yamwong
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * WITTAWAT YAMWONG, OR ANY OTHER CONTRIBUTORS BE LIABLE FOR ANY CLAIM, 
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE 
 * OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 *
 *    Wittawat Yamwong <Wittawat.Yamwong@stud.uni-hannover.de>
 */

/* $Id: mgacontext.h,v 1.1 1999/07/31 07:07:55 crow Exp $ */

#ifndef MGACONTEXT_INC
#define MGACONTEXT_INC

#include "mesaglx/types.h"

#include "mgacommon.h"
#include "mgabuf.h"
#include "mgatex.h"
#include "mgafog.h"
#include "mgavb.h"

#define mgaContextMagic 0x47323030
#define VALID_MGA_CONTEXT(ctx) ((ctx) && (ctx)->magic == mgaContextMagic)

/* for SoftwareFallback */
#define FALLBACK_ALPHATEST 1
#define FALLBACK_BLENDING  2
#define FALLBACK_TEXTURE   4

struct mga_context_t {
  mgaUI32 magic;
  int refcount;

  mgaBufferPtr DB;
  GLcontext *gl_ctx;
  mgaTextureObjectPtr CurrentTexObj;
  mgaTextureObjectPtr DefaultTexObj;
  mgaTextureObjectPtr TexObjList;

  mgaUI32 regDWGCTL;
  mgaUI32 regALPHACTRL;
  mgaUI32 zmode;
  float InvWScale;
  int TextureMode;
  unsigned long current_texture_age;
  struct mga_fog_attrib Fog;

  int SoftwareFallback;  /* or'ed values of FALLBACK_* */

  /* flags */
  int NoZ:1;             /* No z component. */
  int ZNever:1;          /* GL_NEVER = never pass depth test. */
  int NeedAlpha:1;
  int GlobalPaletteUpdated:1; 
  mgaUI16 GlobalPalette[256];
  mgaVertex VB[MGA_VB_SIZE];
};

mgaContextPtr mgaCreateContext(GLcontext *ctx); /* Do I need visual info? */
int mgaDestroyContext(mgaContextPtr ctx);
int mgaBindBuffer(mgaContextPtr ctx, mgaBufferPtr buf);


/*
 * Driver functions that uses mgaCtx or mgaDB and can be called
 * directly from API such as glClear() should use CHECK_CONTEXT
 * before doing anything with mgaCtx and mgaDB.
 *
 * Drawing functions such as Triangle are protected by CHECK_CONTEXT
 * in mgaRenderVB, so that they don't need it.
 *
 * mgaDB will be invalid when the frame buffer have to be reallocate
 * but there was not enough video memory avialable. 
 * TODO: fix this.
 */
#define CHECK_CONTEXT(x) do {					\
  if (!VALID_MGA_CONTEXT(mgaCtx) || !VALID_MGA_BUFFER(mgaDB)) {	\
    x								\
  }								\
} while(0)

#endif
