/*
 * GLX Hardware Device Driver for Matrox G200/G400
 * Copyright (C) 1999 Wittawat Yamwong
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * WITTAWAT YAMWONG, OR ANY OTHER CONTRIBUTORS BE LIABLE FOR ANY CLAIM, 
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE 
 * OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 *
 *    Wittawat Yamwong <Wittawat.Yamwong@stud.uni-hannover.de>
 */

/* $Id: mgadd.c,v 1.3 1999/08/01 04:07:18 johnc Exp $ */

#include "xsmesaP.h"
#include "mesaglx/types.h"
#include "vbrender.h"
#include "glx_log.h"

#include "vga.h"
#include "vgaPCI.h"

#include "mm.h"
#include "g200_mac.h"
#include "mgalib.h"
#include "mgaglx.h"
#include "mgadd.h"
#include "mgadepth.h"
#include "mgatris.h"
#include "mgatex.h"
#include "mgalines.h"
#include "mgalog.h"
#include "mgavb.h"

#ifdef MESA31
#include "vb.h"
#include "dd.h"

#ifdef PROFILE
#  define START_PROFILE                         \
        {                                       \
           GLdouble t0 = gl_time();
 
#  define END_PROFILE( TIMER, COUNTER, INCR )   \
           TIMER += (gl_time() - t0);           \
           COUNTER += INCR;                     \
        }
#else
#  define START_PROFILE
#  define END_PROFILE( TIMER, COUNTER, INCR )
#endif

#endif

extern int MGAydstorg;
extern int MGAmaccess;
extern void xsmesa_setup_DD_pointers( GLcontext *ctx );
extern int xf86VTSema;

#define CHECK_XF86VT(x) if (!xf86VTSema) { x }

static void (*xsmWriteRGBASpan)( const GLcontext *ctx,
				 GLuint n, GLint x, GLint y,
				 CONST GLubyte rgba[][4], 
				 const GLubyte mask[] );
static void (*xsmWriteRGBSpan)( const GLcontext *ctx,
				GLuint n, GLint x, GLint y,
				CONST GLubyte rgb[][3], const GLubyte mask[] );
static void (*xsmWriteMonoRGBASpan)(const GLcontext *ctx, GLuint n, 
				    GLint x, GLint y,const GLubyte mask[] );
static void (*xsmWriteRGBAPixels)(const GLcontext *ctx,
				  GLuint n, const GLint x[], const GLint y[],
				  CONST GLubyte rgba[][4], 
				  const GLubyte mask[] );
static void (*xsmWriteMonoRGBAPixels)(const GLcontext *ctx,GLuint n, 
				      const GLint x[], const GLint y[],
				      const GLubyte mask[] );
static void (*xsmReadRGBASpan)(const GLcontext *ctx, GLuint n, 
			       GLint x, GLint y,GLubyte rgba[][4] );
static void (*xsmReadRGBAPixels)(const GLcontext *ctx,GLuint n, 
				 const GLint x[], const GLint y[],
				 GLubyte rgba[][4], const GLubyte mask[] );

/* Wrapper functions */

static void WriteRGBASpan( const GLcontext *ctx,
			   GLuint n, GLint x, GLint y,
			   CONST GLubyte rgba[][4], 
			   const GLubyte mask[] )
{
  mgaMsg(12,__FUNCTION__ "\n");
  mgaWaitDrawingEngine();
  (*xsmWriteRGBASpan)(ctx,n,x,y,rgba,mask);
}

static void WriteRGBSpan( const GLcontext *ctx,
			  GLuint n, GLint x, GLint y,
			  CONST GLubyte rgb[][3], const GLubyte mask[] )
{
  mgaMsg(12,__FUNCTION__ "\n");
  mgaWaitDrawingEngine();
  (*xsmWriteRGBSpan)(ctx,n,x,y,rgb,mask);
}

static void WriteMonoRGBASpan(const GLcontext *ctx, GLuint n, 
			      GLint x, GLint y,const GLubyte mask[] )
{
  mgaMsg(12,__FUNCTION__ "\n");
  mgaWaitDrawingEngine();
  (*xsmWriteMonoRGBASpan)(ctx,n,x,y,mask);
}

static void WriteRGBAPixels(const GLcontext *ctx,
			    GLuint n, const GLint x[], const GLint y[],
			    CONST GLubyte rgba[][4], 
			    const GLubyte mask[] )
{
  mgaMsg(12,__FUNCTION__ "\n");
  mgaWaitDrawingEngine();
  (*xsmWriteRGBAPixels)(ctx,n,x,y,rgba,mask);
}

static void WriteMonoRGBAPixels(const GLcontext *ctx,GLuint n, 
				const GLint x[], const GLint y[],
				const GLubyte mask[] )
{
  mgaMsg(12,__FUNCTION__ "\n");
  mgaWaitDrawingEngine();
  (*xsmWriteMonoRGBAPixels)(ctx,n,x,y,mask);
}

static void ReadRGBASpan(const GLcontext *ctx, GLuint n, 
			 GLint x, GLint y,GLubyte rgba[][4] )
{
  mgaMsg(12,__FUNCTION__ "\n");
  mgaWaitDrawingEngine();
  (*xsmReadRGBASpan)(ctx,n,x,y,rgba);
}

static void ReadRGBAPixels(const GLcontext *ctx,GLuint n, 
			   const GLint x[], const GLint y[],
			   GLubyte rgba[][4], const GLubyte mask[] )
{
  mgaMsg(12,__FUNCTION__ "\n");
  mgaWaitDrawingEngine();
  (*xsmReadRGBAPixels)(ctx,n,x,y,rgba,mask);
}


/***************************************
 * Mesa's Driver Functions
 ***************************************/
#ifdef MESA31  
static void mgaRenderStart( GLcontext *ctx)
{
  CHECK_XF86VT( return; ); /* !!! FIXME !!! */
  CHECK_CONTEXT( mgaError("mgaDB is not valid in mgaRenderStart().\n");	); /* What to do? */
  mgaMsg(15, __FUNCTION__ "\n");
}

static void mgaRenderFinish( GLcontext *ctx)
{
  mgaMsg(15,__FUNCTION__ "\n");
}


/* Sadly we need the following also: */
/* (yet another macro orgy) */

#define VB_WRAP(x) \
extern void x( struct vertex_buffer *VB,GLuint start,GLuint count,GLuint parity ); 	\
void mga_##x( struct vertex_buffer *VB,GLuint start,GLuint count,GLuint parity ) { \
  CHECK_XF86VT( return; ); 					\
  CHECK_CONTEXT( mgaError("mgaDB is not valid in " #x "\n"););  \
  x(VB, start, count, parity); 					\
  mgaMsg(15, "mga_" #x  " %d %d %d\n", start, count, parity);	\
}

void mga_render_noop( struct vertex_buffer *VB,GLuint start,GLuint count,GLuint parity )
{
  return;
}

#define VB_TAB(y) \
  VB_WRAP(render_vb_points##y)		\
  VB_WRAP(render_vb_lines##y)		\
  VB_WRAP(render_vb_line_loop##y)	\
  VB_WRAP(render_vb_line_strip##y)	\
  VB_WRAP(render_vb_triangles##y)	\
  VB_WRAP(render_vb_tri_strip##y)	\
  VB_WRAP(render_vb_tri_fan##y)		\
  VB_WRAP(render_vb_poly##y)		\
  VB_WRAP(render_vb_quads##y)		\
  VB_WRAP(render_vb_quad_strip##y)	\
  render_func mga_render_tab##y[] = {	\
    mga_render_vb_points##y,		\
    mga_render_vb_lines##y,		\
    mga_render_vb_line_loop##y,		\
    mga_render_vb_line_strip##y,	\
    mga_render_vb_triangles##y,		\
    mga_render_vb_tri_strip##y,		\
    mga_render_vb_tri_fan##y,		\
    mga_render_vb_quads##y,		\
    mga_render_vb_quad_strip##y,	\
    mga_render_vb_poly##y,		\
    mga_render_noop };
   
VB_TAB(_raw)
VB_TAB(_cull)
VB_TAB(_clipped)

#else

static GLboolean mgaRenderVB( GLcontext *ctx, GLboolean allDone )
{
  CHECK_XF86VT( gl_reset_vb(ctx,allDone); return GL_TRUE; );
  CHECK_CONTEXT(					\
    mgaError("mgaDB is not valid in mgaRenderVB().\n");	\
    gl_reset_vb(ctx,allDone);				\
    return GL_TRUE;					\
  );

  mgaMsg(15, "mgaRenderVB()\n");
  gl_render_vb(ctx,allDone);
  return GL_TRUE;
}
#endif /* MESA31 */

/*
 * mgaClear
 * perform hardware accelerated clearing of the color and/or depth
 * buffer.  Software may still clear stencil buffers.
 * If all==GL_TRUE, clear whole buffer, else just clear region defined
 * by x,y,width,height
 */
 #define MGA_CLEAR_COLOR 1
#define MGA_CLEAR_Z     2
#define MGA_CLEAR_ALL   MGA_CLEAR_COLOR|MGA_CLEAR_Z


static GLbitfield mgaClear( GLcontext *ctx, GLbitfield mask, GLboolean all,
			 GLint x, GLint y, GLint width, GLint height ) {
	XSMesaContext xsmesa = (XSMesaContext) ctx->DriverCtx;
	GLubyte	*c;
	mgaUI32	zval;
	DMALOCALS;
	mgaUI32	cmd;

	CHECK_XF86VT ( return mask; );
	CHECK_CONTEXT(					\
		mgaError("mgaDB is not valid in mgaClear().\n");	\
		return mask;					\
	);
	if ( !mgaDB->Drawable ) {
		return mask;
	}

	if (all == GL_TRUE) {
		x = 0;
		y = 0;
		width = mgaDB->Width;
		height = mgaDB->Height;
	}
	
	/* flip top to bottom */
	y = mgaDB->Height-y-height;
	c = xsmesa->clearcolor;
	zval = (mgaUI32) (ctx->Depth.Clear * DEPTH_SCALE);
	
	/* add the clear commands to the dma buffer */
	/*	ENTER3D; */
	if ( mgaglx.needEnter3D ) {
		if ( !mgaSet3DState() ) { 
			return mask; /* can't do 3D drawing for some reason */
		}
	}

	MGADMAGETPTR( 32 );

	cmd = DC_opcod_trap | DC_arzero_enable | DC_sgnzero_enable | 
	  DC_shftzero_enable | (0xC << DC_bop_SHIFT) | DC_clipdis_enable | 
	  DC_atype_rstr | DC_solid_enable | DC_transc_enable;

	/* color buffer */
	if ( mask & GL_COLOR_BUFFER_BIT ) {
		mask &= ~GL_COLOR_BUFFER_BIT;
	
		MGADMA_FCOL(mgaPackColor(c[0],c[1],c[2],c[3]));
		MGADMA_YDSTLEN(y,height);
		MGADMA_FXBNDRY(x,x+width);
		MGADMA_DWGCTL_EXEC(cmd);
	}

	/* depth buffer */
	if ( (mask & GL_DEPTH_BUFFER_BIT) && mgaDB->HasZORG && ctx->Depth.Mask ) {
		mgaUI32 maccess = mgaDB->Setup[MGA_SETUP_MACCESS];
		
		mask &= ~GL_DEPTH_BUFFER_BIT;

		DMAOUTREG(MGAREG_DSTORG,mgaDB->ZBuffer->Setup[MGA_SETUP_ZORG]);
		#if DEPTH_BITS==16
		MGA_SET_FIELD(maccess,MA_pwidth_MASK,MA_pwidth_16);
		#elif DEPTH_BITS==32
		MGA_SET_FIELD(maccess,MA_pwidth_MASK,MA_pwidth_32);
		#else
		#  error DEPTH_BITS must be either 16 or 32.
		#endif
		DMAOUTREG(MGAREG_MACCESS,maccess);
		MGADMA_FCOL(zval);
		MGADMA_YDSTLEN(y,height);
		MGADMA_FXBNDRY(x,x+width);
		MGADMA_DWGCTL_EXEC(cmd);
		DMAOUTREG(MGAREG_DSTORG,mgaDB->Setup[MGA_SETUP_DSTORG]);
		DMAOUTREG(MGAREG_MACCESS,mgaDB->Setup[MGA_SETUP_MACCESS]);
	}

	DMAADVANCE();

	/* we are NOT flushing the dma here.  if things are running smoothly,
	the only dma flush will be at swapbuffers time */

	return mask;
}

static void mgaNearFar( GLcontext *ctx, GLfloat nearVal, GLfloat farVal )
{
  CHECK_CONTEXT( return; );

#define INVWMAX 128.0F

  mgaCtx->InvWScale = 1.0;
  if (nearVal > 0) {
    /* limit InvWScale/wMin in (0,INVWMAX] to avoid over- and underflow. 
       InvWScale is used by texture setup in mga_tritemp.h */
    int exp2;    
    if (frexp(INVWMAX * nearVal,&exp2) != 0) {
      if (exp2 >= 2) {
        mgaCtx->InvWScale = 1 << (exp2-1);
      } else if (exp2 <= 0) {
        mgaCtx->InvWScale = 1.0 / (1 << (-exp2+1));
      }          
    }            
  }
  mgaMsg(11,"mgaNearFar(): nearVal=%f, InvWScale=%f\n",nearVal,mgaCtx->InvWScale);
}

static GLint mgaGetParameteri(const GLcontext *ctx, GLint param)
{
  switch (param) {
#ifndef MESA31    
  case DD_MAX_TEXTURE_SIZE:
    return 1024;
  case DD_MAX_TEXTURES:
      return 1;
  case DD_MAX_TEXTURE_COORD_SETS:
    return MAX_TEX_COORD_SETS;
#endif    
  case DD_HAVE_HARDWARE_FOG:  
    return 0; /* actually we have one but it doesn't work well. */
  default:
    mgaError("mgaGetParameteri(): unknown parameter!\n");
    return 0; /* Should I really return 0? */
  }
}

static void mgaUpdateZMode(const GLcontext *ctx)
{
  int zmode = 0;

  mgaCtx->ZNever = 0;

  if (ctx->Depth.Test && mgaDB->HasZORG) {
    mgaCtx->NoZ = 0;
    switch(ctx->Depth.Func)  { 
    case GL_NEVER:
      zmode = DC_zmode_nozcmp; mgaCtx->ZNever = 1; break;
    case GL_ALWAYS:  
      zmode = DC_zmode_nozcmp; break;
    case GL_LESS:
      zmode = DC_zmode_zlt; break; 
    case GL_LEQUAL:
      zmode = DC_zmode_zlte; break;
    case GL_EQUAL:
      zmode = DC_zmode_ze; break;
    case GL_GREATER:
      zmode = DC_zmode_zgt; break;
    case GL_GEQUAL:
      zmode = DC_zmode_zgte; break;
    case GL_NOTEQUAL:
      zmode = DC_zmode_zne; break;
    default:
      break;
    }
    if (ctx->Depth.Mask)
      mgaCtx->zmode = zmode | DC_atype_zi;
    else
      mgaCtx->zmode = zmode | DC_atype_i;
  } else {
    mgaCtx->NoZ = 1;
    mgaCtx->zmode = DC_zmode_nozcmp | DC_atype_i;
  }
}

static void mgaUpdateAlphaMode(const GLcontext *ctx)
{
  int needdst = 0;
  int a = 0;

  /* disable alpha test/blending as default */
  mgaCtx->regALPHACTRL = AC_src_one | AC_amode_alpha_channel;
  mgaCtx->NeedAlpha = 0;
  if (ctx->Color.AlphaEnabled) {
    mgaCtx->NeedAlpha = 1;
    switch (ctx->Color.AlphaFunc) {
    case GL_LESS:
      a |= AC_atmode_alt; break;
    case GL_GEQUAL:
      a |= AC_atmode_agte; break;
    case GL_LEQUAL:  
      a |= AC_atmode_alte; break;
    case GL_GREATER:  
      a |= AC_atmode_agt; break;
    case GL_NOTEQUAL: 
      a |= AC_atmode_ane; break;
    case GL_EQUAL: 
      a |= AC_atmode_ae; break;
    case GL_ALWAYS:
      a |= AC_atmode_noacmp; break;
    default:
      mgaCtx->SoftwareFallback |= FALLBACK_ALPHATEST;
      return;
    }
    a |= MGA_FIELD(AC_atref,ctx->Color.AlphaRef);
  }
  if (ctx->Color.BlendEnabled) {
    mgaCtx->NeedAlpha = 1;
    if (ctx->Color.BlendEquation != GL_FUNC_ADD_EXT) {
      mgaCtx->SoftwareFallback |= FALLBACK_BLENDING;
      return;
    }
#ifdef MESA31
    switch (ctx->Color.BlendSrcRGB) {
#endif
#ifndef MESA31
    switch (ctx->Color.BlendSrc) {
#endif
      case GL_ZERO:
	a |= AC_src_zero; break;
      case GL_SRC_ALPHA:
	a |= AC_src_src_alpha; break;
      case GL_ONE:
	a |= AC_src_one; break;
      case GL_DST_COLOR:
	a |= AC_src_dst_color; break;
      case GL_ONE_MINUS_DST_COLOR:
	a |= AC_src_om_dst_color; break;
      case GL_ONE_MINUS_SRC_ALPHA:
	a |= AC_src_om_src_alpha; break;
      case GL_DST_ALPHA:
	a |= AC_src_dst_alpha;
	needdst = 1; 
	break;
      case GL_ONE_MINUS_DST_ALPHA:
	a |= AC_src_om_dst_alpha; 
	needdst = 1;
	break;
      case GL_SRC_ALPHA_SATURATE:
	a |= AC_src_src_alpha_sat; break;
    default:
      mgaCtx->SoftwareFallback |= FALLBACK_BLENDING;
      return;
    }
#ifdef MESA31
    switch (ctx->Color.BlendDstRGB) {
#endif
#ifndef MESA31
    switch (ctx->Color.BlendDst) {
#endif
      case GL_SRC_ALPHA:
	a |= AC_dst_src_alpha; break;
      case GL_ONE_MINUS_SRC_ALPHA:
	a |= AC_dst_om_src_alpha; break;
      case GL_ZERO:
	a |= AC_dst_zero; break;
      case GL_ONE:
	a |= AC_dst_one; break;
      case GL_SRC_COLOR:
	a |= AC_dst_src_color; break;
      case GL_ONE_MINUS_SRC_COLOR:
	a |= AC_dst_om_src_color; break;
      case GL_DST_ALPHA:
	a |= AC_dst_dst_alpha;
	needdst = 1;
	break;
      case GL_ONE_MINUS_DST_ALPHA:
	a |= AC_dst_om_dst_alpha;
	needdst = 1;
	break;
    default:
      mgaCtx->SoftwareFallback |= FALLBACK_BLENDING;
      return;
    }
    if (!(mgaDB->Attrib & MGA_PF_HASALPHA) && needdst) {
      mgaCtx->SoftwareFallback |= FALLBACK_BLENDING;
      return;
    }
  } else {
    MGA_SET_FIELD(a,AC_src_MASK & AC_dst_MASK,
		  AC_src_one | AC_dst_zero) ; /* disable blending */
  }

  if (ctx->Texture.Enabled && mgaCtx->CurrentTexObj) {
    if ((mgaCtx->CurrentTexObj->TexReg[MGA_TEXREG_TEXCTL] & ~TMC_tformat_MASK)
        == TMC_tformat_tw16) {
      a |= AC_alphasel_diffused;
    } else {
      switch (mgaCtx->TextureMode) {
      case GL_DECAL:    a |= AC_alphasel_diffused; break;
      case GL_MODULATE: a |= AC_alphasel_modulated; break;
      default:  /* a |= AC_alphasel_fromtex; */ break;
      }
    }
  }

  mgaCtx->regALPHACTRL = a | AC_amode_alpha_channel;
  mgaCtx->SoftwareFallback &= ~(FALLBACK_ALPHATEST | FALLBACK_BLENDING);
}


static void mgaUpdateClipping(const GLcontext *ctx)
{
  int x1,x2,y1,y2;
  mgaUI32 tmp;

  if (ctx->Scissor.Enabled) {
    x1 = ctx->Scissor.X;
    x2 = ctx->Scissor.X + ctx->Scissor.Width - 1;
    y1 = mgaDB->Height - ctx->Scissor.Y - ctx->Scissor.Height;
    y2 = mgaDB->Height - ctx->Scissor.Y - 1;
  } else {
  	x1 = 0;
  	y1 = 0;
  	x2 = mgaDB->Width-1;
  	y2 = mgaDB->Height-1;
  }
  
  if (x1 < 0) x1 = 0;
  if (y1 < 0) y1 = 0;
  if (x2 >= mgaDB->Width) x2 = mgaDB->Width-1;
  if (y2 >= mgaDB->Height) y2 = mgaDB->Height-1;
  if (x1 > x2 || y1 > y2) {
    /* this will set clipping so that nothing will be drawn. */
    x1 = 0; x2 = 0;
    y2 = 0; y1 = 1;
  }
  tmp = MGA_FIELD(CXB_cxright,x2) | MGA_FIELD(CXB_cxleft,x1);
  if (mgaDB->Setup[MGA_SETUP_CXBNDRY] != tmp) {
    mgaDB->Setup[MGA_SETUP_CXBNDRY] = tmp;
  }
  tmp = y1*mgaDB->Pitch;
  if (mgaDB->Setup[MGA_SETUP_YTOP] != tmp) {
    mgaDB->Setup[MGA_SETUP_YTOP] = tmp;
  }
  tmp = y2*mgaDB->Pitch;
  if (mgaDB->Setup[MGA_SETUP_YBOT] != tmp) {
    mgaDB->Setup[MGA_SETUP_YBOT] = tmp;
  }
  mgaCtx->regDWGCTL &= ~DC_clipdis_enable;
}

static void mgaUpdateColorMask(const GLcontext *ctx)
{
  GLuint mask;
  
  mask = mgaPackColor(ctx->Color.ColorMask[RCOMP],
                      ctx->Color.ColorMask[GCOMP],
                      ctx->Color.ColorMask[BCOMP],
		      ctx->Color.ColorMask[ACOMP]);
  
  switch (mgaDB->Attrib & MGA_PF_MASK)
  {
    case MGA_PF_555:
    case MGA_PF_565:
      mask = mask | (mask << 16);
      break;
  }
  mgaDB->Setup[MGA_SETUP_PLNWT] = mask;
}

extern struct gl_texture_object *HashLookup(void *to, int name);
GLboolean mgaAreTexturesResident( GLcontext *ctx, GLsizei n,
                                  const GLuint *texName,
                                  GLboolean *residences )
{
   GLboolean resident = GL_TRUE;
   GLint i;

   if (n<0) {
      mgaError( "mgaAreTexturesResident(n)" );
      return GL_FALSE;
   }

   for (i=0;i<n;i++) {
      struct gl_texture_object *t;
      if (texName[i]==0) {
         mgaError("glAreTexturesResident(textures)" );
         return GL_FALSE;
      }
      t = HashLookup(ctx->Shared->TexObjects, texName[i]);
      if (t) {
	 mgaTextureObjectPtr mt =( mgaTextureObjectPtr) t->DriverData;
	residences[i] = (mt && (!mt->swapped_out))?GL_TRUE:GL_FALSE;
	resident = resident && residences[i];
      }
      else {
         mgaError( "glAreTexturesResident(textures)" );
         return GL_FALSE;
      }
   }
   return resident;
}

const char *mgaRendererString()
{
    /*
    * Return a string which uniquely identifies this device driver.
    * The string should contain no whitespace.  Examples: "X11" "OffScreen"
    * "MSWindows" "SVGA".
    */
  return "MGA-GLX";
}

const char *mgaExtensionString(GLcontext *ctx)
{
   /* Return a space-separated list of extensions for this driver.
    */
  static char *mgaOpenGLExtensions = 
     "GL_EXT_vertex_array GL_EXT_texture_object " 
     "GL_EXT_abgr GL_EXT_texture "
     "GL_EXT_blend_color GL_EXT_blend_logic_op "
     "GL_EXT_paletted_texture GL_EXT_shared_texture_palette"
    ;
  /* Any more supported implicitly through Mesa (and GLX protocol) yet? */

  return mgaOpenGLExtensions;
}

static void mga_null_triangle( GLcontext *ctx, GLuint v0,
                          GLuint v1, GLuint v2, GLuint pv ) {
}     
static void mga_null_line( GLcontext *ctx, GLuint v1, GLuint v2, GLuint pv ) {
}

static void mga_null_points( GLcontext *ctx, GLuint first, GLuint last ) {
}

static int mgaCanUseHardware(const GLcontext *ctx)
{
  XSMesaContext xsmesa = (XSMesaContext) ctx->DriverCtx;
  XSMesaBuffer xsmbuf = xsmesa->xsm_buffer;
  int allvalid = (mgaDB && mgaDB->Drawable && xsmbuf->buffer == XIMAGE);

  CHECK_CONTEXT( return 0; );

  if ( !allvalid )
    mgaMsg(10, "DB:%p Drawable:%d XIMAGE:%d\n", mgaDB ,mgaDB->Drawable, (xsmbuf->buffer == XIMAGE));
  return allvalid;
}

#define WRAP(x) 				\
    xsm ## x = ctx->Driver.x;			\
    ctx->Driver.x = x

/*
 * if we can't hardware accelerate the context now, but could
 * before (the window was sized up to the point a buffer alloc failed)
 * then just have software mesa do everything.
 */
static void mga_setup_DD_pointers_no_accel( GLcontext *ctx ) {
    /* disable hardware accel. */
    mgaMsg(5,"no hw accel.\n");
    /* Do I really have to clear these pointer? */
    ctx->Driver.TriangleFunc = NULL;
    ctx->Driver.LineFunc = NULL;
#ifdef MESA31
    ctx->Driver.RenderStart  = NULL; 
    ctx->Driver.RenderFinish = NULL; 
    ctx->Driver.RenderVBRawTab = NULL;
    ctx->Driver.RenderVBCulledTab = NULL;
    ctx->Driver.RenderVBClippedTab = NULL;
#else
    ctx->Driver.RenderVB = NULL;
#endif    
    ctx->Driver.NearFar = NULL;
    ctx->Driver.GetParameteri = NULL;
    ctx->Driver.RasterSetup = NULL;
    
    ctx->Driver.ExtensionString = NULL;

    ctx->Driver.TexEnv = NULL;
    ctx->Driver.TexImage = NULL;
    ctx->Driver.BindTexture = NULL;
    ctx->Driver.DeleteTexture = NULL;
    ctx->Driver.TexParameter = NULL;
    
    xsmesa_setup_DD_pointers(ctx);
    ctx->Driver.UpdateState = mga_setup_DD_pointers;
}

points_func mga_choose_points_function( GLcontext *ctx );

/*
 * mga_setup_DD_pointers
 * mesa calls this before drawing any primitives if the state
 * has changed.  State changes are batched up, this will not be
 * called for every glEnable/glDisable.
 * This does not generate ANY dma command traffic, but it
 * sets a flag to force the state to be updated before the
 * next primitive is drawn.  This prevents setting state for
 * primitves that get clipped away.
 * TODO: split into 2 parts: First time setup and update function 
 */
void mga_setup_DD_pointers( GLcontext *ctx ) {
	mgaglx.c_setupPointers++;

	mgaMsg( 15, "mga_setup_DD_pointers( %p )\n", ctx );

	if ( ctx != mgaCtx->gl_ctx ) {
		mgaMsg( 1, "mga_setup_DD_pointers: ctx != mgaCtx->gl_ctx\n" );
		mga_setup_DD_pointers_no_accel( ctx );
		return;
	}
	
	if ( !mgaCanUseHardware(ctx) ) {
		mga_setup_DD_pointers_no_accel( ctx );
		return;
	}
  
	xsmesa_setup_DD_pointers(ctx);
	ctx->Driver.UpdateState = mga_setup_DD_pointers;

	mgaUpdateZMode(ctx);
	mgaUpdateAlphaMode(ctx);
	mgaUpdateFogAttrib(mgaCtx);
	mgaUpdateClipping(ctx);
	mgaUpdateColorMask(ctx);

	/* set the texturing variables */
	mgaUpdateTextureState( ctx, mgaCtx->CurrentTexObj );

 	/* we will need to update the drawing registers before we
 	draw our next primitive */
	mgaglx.needEnter3D = 1; 

    WRAP(WriteRGBASpan);
    WRAP(WriteRGBSpan);
    WRAP(WriteMonoRGBASpan);
    WRAP(WriteRGBAPixels);
    WRAP(WriteMonoRGBAPixels);
    WRAP(ReadRGBASpan);
    WRAP(ReadRGBAPixels);

    ctx->Driver.RendererString  = mgaRendererString;
    ctx->Driver.ExtensionString = mgaExtensionString;
    
#ifdef MESA31
    ctx->Driver.RenderStart  = mgaRenderStart; 
    ctx->Driver.RenderFinish = mgaRenderFinish; 
    ctx->Driver.RenderVBRawTab = mga_render_tab_raw;
    ctx->Driver.RenderVBCulledTab = mga_render_tab_cull;
    ctx->Driver.RenderVBClippedTab = mga_render_tab_clipped;
    
    ctx->Driver.TriangleCaps = DD_TRI_LIGHT_TWOSIDE;
#else
    ctx->Driver.RenderVB = mgaRenderVB;
#endif    
    ctx->Driver.Clear = mgaClear;
    ctx->Driver.NearFar = mgaNearFar;
    ctx->Driver.GetParameteri = mgaGetParameteri;
    ctx->Driver.RasterSetup = mgaChooseRasterSetupFunc(ctx);

    ctx->Driver.TexEnv = mgaTexEnv;
    ctx->Driver.TexImage = mgaTexImage;
    ctx->Driver.BindTexture = mgaBindTexture;
    ctx->Driver.DeleteTexture = mgaDeleteTexture;
    ctx->Driver.TexParameter = mgaTexParameter;
    ctx->Driver.UpdateTexturePalette = mgaUpdateTexturePalette;
/* 
 * we dont need  ctx->Driver.UseGlobalTexturePalette 
 * because we can use the flag set by sw mesa.
 */
    
    /* There is no driver callback so I need to manipulate the API struct... 
     * !!! should be changed in Mesa !!! (is not in 3.x)    - Andree */
    ctx->API.AreTexturesResident = mgaAreTexturesResident;

    if (mgaDB->HasZORG) {
      ctx->Driver.ReadDepthSpanFloat = mga_read_depth_span_float;
      ctx->Driver.ReadDepthSpanInt = mga_read_depth_span_int;
      if (ctx->Depth.Test) {
        switch (ctx->Depth.Func) {
        case GL_LESS:
  	  ctx->Driver.DepthTestSpan = mga_depth_test_span_less;
  	  ctx->Driver.DepthTestPixels = mga_depth_test_pixels_less;
  	  break;
        case GL_GREATER:
  	  ctx->Driver.DepthTestSpan = mga_depth_test_span_greater;
  	  ctx->Driver.DepthTestPixels = mga_depth_test_pixels_greater;
  	  break;
        default:
  	  ctx->Driver.DepthTestSpan = mga_depth_test_span_generic;
  	  ctx->Driver.DepthTestPixels = mga_depth_test_pixels_generic;
  	  break;
        }
      }
    }

    if (mgaCtx->ZNever || /* never pass depth test, so draw nothing */
	(mgaCtx->SoftwareFallback != 0)) {
	  ctx->Driver.TriangleFunc = NULL;
	  ctx->Driver.LineFunc = NULL;
	  ctx->Driver.PointsFunc = NULL;
	  if (mgaCtx->SoftwareFallback) {
		mgaMsg(5, "software fallback! (%d)\n", mgaCtx->SoftwareFallback);
	  }
    } else {
      if ( mgaglx.nullprims ) {	/* performance testing */
	      ctx->Driver.TriangleFunc = mga_null_triangle; 
	      ctx->Driver.LineFunc = mga_null_line;
	      ctx->Driver.PointsFunc = mga_null_points;
	} else {
	      ctx->Driver.TriangleFunc = mga_choose_triangle_function(ctx);  
	      ctx->Driver.LineFunc = mga_choose_line_function(ctx);
	      ctx->Driver.PointsFunc = mga_choose_points_function(ctx);
	}
    }
}
