/*
 * GLX Hardware Device Driver for Matrox G200/G400
 * Copyright (C) 1999 Wittawat Yamwong
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * WITTAWAT YAMWONG, OR ANY OTHER CONTRIBUTORS BE LIABLE FOR ANY CLAIM, 
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE 
 * OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 *
 *    Wittawat Yamwong <Wittawat.Yamwong@stud.uni-hannover.de>
 */

/* $Id: mgaglx.c,v 1.2 1999/08/01 04:07:18 johnc Exp $ */

#include <stdlib.h>

#include "context.h"
#include "depth.h"
#include "macros.h"
#include "texstate.h"
#include "triangle.h"
#include "vb.h"
#include "types.h"

#include "xsmesaP.h"
#include "glx_log.h"
#include "mesaglx/context.h"
#include "mesaglx/matrix.h"
#include "mesaglx/types.h"

#define GC XXGC
#include "gcstruct.h"
#include "pixmapstr.h"
#include "servermd.h" /* PixmapBytePad */
#include "scrnintstr.h"
#include "regionstr.h"
#include "windowstr.h"
#undef GC

#include "vga.h"
#include "vgaPCI.h"
#include "mga.h"
#include "xaa/xf86xaa.h"

#include "mm.h"
#include "mgabuf.h"
#include "mgadd.h"
#include "g200_mac.h"
#include "mgalib.h"
#include "mgalog.h"

#ifndef NO_MTRR
#define MTRR_NEED_STRINGS
#include <errno.h>
#include <asm/mtrr.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <unistd.h>
int mtrr;
void mgaCloseMTRR();
#endif

extern int xf86VTSema;
extern XSMesaContext XSMesa;

/* our collected globals */
mgaGlx_t	mgaglx;

GLboolean mgaInitGLX();

/* are these for the multi-hardware glx support? */
GLboolean (*hwInitGLX)() = mgaInitGLX;
GLboolean (*hwInitVisuals)(VisualPtr *, DepthPtr *, int *, int *, 
                           int *, VisualID *, unsigned long, int) = NULL;

void mgaDumpDB(mgaBufferPtr buf)
{
/*  int i;  */

  mmDumpMemInfo( cardHeap );
  mgaMsg(1,"Dump DB:\n");
  mgaMsg(1,"  %s %s\n",(buf->Drawable) ? "Drawable" : "-",
         (buf->HasZORG) ? "HasZORG" : "-");
  /*
  for (i=0; i < sizeof(buf->Setup)/sizeof(mgaUI32); i++)
    mgaMsg(1,"  %d: 0x%08x\n",i,buf->Setup[i]);
  */
  mgaMsg(1,"  w,h,p = %d,%d,%d\n",buf->Width,buf->Height,buf->Pitch);
  mgaMsg(1,"End Dump DB\n");
}

void mgaGLXCreateDepthBuffer(GLcontext* ctx)
{
  XSMesaContext xsmesa = (XSMesaContext) ctx->DriverCtx;
  mgaBufferPtr buf,zb=0;

	mgaMsg( 1, "mgaGLXCreateDepthBuffer\n" );
	
  /* deallocate current depth buffer if present */
  if (ctx->Buffer->Depth)
    {
      free(ctx->Buffer->Depth);
      ctx->Buffer->Depth = NULL;
    }
  if (xsmesa->xsm_buffer->db_state == BACK_XIMAGE &&
      (buf = (mgaBufferPtr)xsmesa->xsm_buffer->backimage->devPriv) != 0) {
#if DEPTH_BITS==16
    zb = mgaAttachZBuffer(buf,NULL,MGA_ZBUFFER16);
#elif DEPTH_BITS==32
    zb = mgaAttachZBuffer(buf,NULL,MGA_ZBUFFER32);
#endif
    if (!zb) {
      /* We need a depth buffer, but cannot allocate it. So, disable hardware
	 acceleration. */
      buf->Drawable = 0;
      mgaError("mgaGLXCreateDepthBuffer(): mgaAttachZBuffer() failed!\n");
    }
    if (mgaLogLevel(1))  {
    	mgaDumpDB(buf);
    }
  }
  /* allocate new depth buffer, but don't initialize it */
  /* FIXME: we shouldn't need to do this if we got a hardware buffer */
  ctx->Buffer->Depth = (GLdepth *) malloc(ctx->Buffer->Width
					  * ctx->Buffer->Height
					  * sizeof(GLdepth));
  if (!ctx->Buffer->Depth)
    {
      /* out of memory */
      ctx->Depth.Test = GL_FALSE;
      mgaError("Couldn't allocate depth buffer\n" );
    }

}

/*
 * CreateFrontBuffer
 * Called during initialization to set up parameters needed
 * for the swapbuffers blit.
 */
static mgaBufferPtr CreateFrontBuffer( void )
{
  int Attrib;
  mgaBufferPtr buf;
  
  switch( vga256InfoRec.depth )  {
  case 15:
    Attrib = MGA_PF_555;
    break;
  case 16:
    Attrib = MGA_PF_565;
    break;
  case 24:
    if (vgaBitsPerPixel == 24)
      Attrib = MGA_PF_888;	/* we can't render to this, but we can blit to it */
    else
      Attrib = MGA_PF_8888;
    break;
  default:
    mgaError("No support for %d bits per pixel.\n",vgaBitsPerPixel);
    return NULL;
  }
  buf = mgaCreatePrimaryBuffer(Attrib,
			       vga256InfoRec.virtualX,
			       vga256InfoRec.virtualY,
			       vga256InfoRec.displayWidth);

  /* enable hardware dithering in 16 bit modes */
  switch(Attrib & MGA_PF_MASK) {
  case MGA_PF_555:
  case MGA_PF_565:
    buf->Setup[MGA_SETUP_MACCESS] |= MA_dit555_enable | 
      MA_nodither_enable; /* xf86 default */
    break;
  }
  buf->Setup[MGA_SETUP_CXBNDRY] = 0x0fff0000;
  buf->Setup[MGA_SETUP_YTOP] = 0;
  buf->Setup[MGA_SETUP_YBOT] = 0x00ffffff;
  return buf;
}


void mgaGLXDestroyImage(GLXImage* image)
{
  if (image->devPriv) {
    mgaDestroyBuffer((mgaBufferPtr)image->devPriv);
    if (mgaLogLevel(10)) {
       mmDumpMemInfo( cardHeap );
     }
  } else if (image->data)
    free(image->data);
  xfree(image);
}

/*
 * mgaGLXCreateImage
 * We can't accelerate rendering to images
 */
GLXImage* mgaGLXCreateImage(WindowPtr pwindow, int depth, int width, int height)
{
  int Attrib;
  mgaBufferPtr buf;
  GLXImage* image=(GLXImage*)xalloc(sizeof(GLXImage));

  if (!image)
    return (NULL);
  image->pwin = pwindow;
  image->width=width;
  image->height=height;
  image->bits_per_pixel=depth;
  image->data=0;
  image->devPriv=NULL;

  Attrib = -1;
  switch (depth)
    {
    case 8:
      Attrib = MGA_PF_INDEX;
      break;
    case 15:
      Attrib = MGA_PF_555;
      break;
    case 16:
      Attrib = MGA_PF_565;
      break;
    case 24:	/* we can't render into a 24 bit image, so use a 32 bit one */
    case 32:
      Attrib = MGA_PF_8888;
      break;
    default:
      mgaError("Unknown width in GLXCreateImage\n");
      break;
    }
  if (Attrib != -1)
    buf = mgaCreateBuffer(Attrib | MGA_COLORBUFFER,0,width,height,0);
  else
    buf = NULL;
  image->devPriv = (void *) buf;
  if (!image->devPriv) {
    image->bytes_per_line=PixmapBytePad(width, depth);
    image->data = (char *) malloc( image->height * image->bytes_per_line );
    if (!image->data)
      {
	mgaError("mgaGLXCreateImage: malloc failed.");
	xfree(image);
	image = NULL;
      }
  } else {
    buf->refcount++;
    image->bytes_per_line=buf->Pitch * buf->BytesPerPixel;
    /* How can I draw XImage with any line pad? */
    /* e.g. bytes_per_line != PixmapBytePad */
    /* My work around: adjust width instead */
    image->width = buf->Pitch;
    image->data = buf->BufAddr;
    if (mgaLogLevel(1)) {
       mmDumpMemInfo( cardHeap );
     }
  }
  return (image);
}


/*
 * BackToFront
 * Hardware accelerated back buffer blit, properly clipped to
 * all the X windows
 */
static void BackToFront(DrawablePtr drawable, mgaBufferPtr buf) {
	RegionPtr		prgnClip;
	BoxPtr		pbox;
	int 			i,nbox;
	int 			xorg, yorg, pitch;
	DMALOCALS;

	if (!VALID_MGA_BUFFER(buf)) {
		mgaError("BackToFront(): invalid back buffer\n");
		return;
	}
	if ( !VALID_MGA_BUFFER(mgaFrontBuffer) ) {
		mgaError("BackToFront(): invalid mgaFrontBuffer\n");
		return;
	}
	if ( !xf86VTSema ) {
		mgaError("BackToFront(): !xf86VTSema\n");
		return;
	}
	if ( drawable->width != buf->Width 
		|| drawable->height != buf->Height 
	 	|| drawable->type != DRAWABLE_WINDOW ) {
		mgaError("BackToFront(): bad drawable\n");
		return;
	}

	prgnClip = &((WindowPtr)drawable)->clipList;
	pbox = REGION_RECTS(prgnClip);
	nbox = REGION_NUM_RECTS(prgnClip);
	if( !nbox ) {
		/* window is completely covered */
		return;
	}

	xorg = drawable->x;
	yorg = drawable->y;
	pitch = buf->Pitch;

	/* get out of 3D mode for the blit */
	mgaDmaExecute(mgaFrontBuffer->Setup,mgaFrontBuffer->SetupSize);
	
	/* we will need to go back to setup the back buffer before drawing again */
	mgaglx.needEnter3D = 1;

	/* add the blit commands to the dma buffer */
	MGADMAGETPTR( 10 );

	MGADMA_DSTORG(mgaFrontBuffer->Setup[MGA_SETUP_DSTORG]);
	MGADMA_SRCORG(buf->Setup[MGA_SETUP_DSTORG]);
	DMAOUTREG(MGAREG_AR5,pitch);

	if ( vgaBitsPerPixel == 24 ) { 
	/* FIXME: we can't use the blitter to copy a 32 bit back buffer to a 24 bit
	screen, so we will have to do this with the CPU when we get around to it... */
		MGADMA_DWGCTL(DC_opcod_iload | DC_atype_rstr | DC_linear_xy |
		     DC_solid_disable | DC_arzero_disable | DC_sgnzero_enable |
		     DC_shftzero_enable | (0xC << DC_bop_SHIFT) | DC_bltmod_bu32rgb |
		     DC_pattern_disable | DC_transc_disable | DC_clipdis_disable); 
	} else {
		MGADMA_DWGCTL(DC_opcod_bitblt | DC_atype_rpl | DC_linear_xy |
		     DC_solid_disable | DC_arzero_disable | DC_sgnzero_enable |
		     DC_shftzero_enable | (0xC << DC_bop_SHIFT) | DC_bltmod_bfcol |
		     DC_pattern_disable | DC_transc_disable | DC_clipdis_disable); 
	}	     
	DMAADVANCE();
	     
	/* xy bitblt each visible rectangle */
	for (i=nbox; i > 0; i--) {
		int x = pbox->x1 - xorg;
		int y = pbox->y1 - yorg;
		int w = pbox->x2 - pbox->x1;
		int h = pbox->y2 - pbox->y1;
		int start = x + y*pitch;

		MGADMAGETPTR( 10 );

		DMAOUTREG(MGAREG_AR0,start + w - 1);
		DMAOUTREG(MGAREG_AR3,start);
		MGADMA_FXBNDRY(pbox->x1,pbox->x2-1);
		MGADMA_YDSTLEN_EXEC(pbox->y1,h);
		pbox++;

		DMAADVANCE();
	}
  
   	MGADMAGETPTR( 10 );
	MGADMA_SRCORG(0);
 	DMAADVANCE();
    
}

/*
 * ClearBox
 * Add hardware commands to draw a filled box for the
 * debugging display.
 */
static void ClearBox( int x, int y, int w, int h, int r, int g, int b ) {
	DMALOCALS;
	mgaUI32	cmd;

	MGADMAGETPTR( 32 );

	cmd = DC_opcod_trap | DC_arzero_enable | DC_sgnzero_enable | 
	  DC_shftzero_enable | (0xC << DC_bop_SHIFT)  | 
	  DC_atype_rstr | DC_solid_enable;

	MGADMA_FCOL(mgaPackColor(r,g,b,255));
	MGADMA_YDSTLEN(y,h);
	MGADMA_FXBNDRY(x,x+w);
	MGADMA_DWGCTL_EXEC(cmd);

	DMAADVANCE();
}

/*
 * performanceBoxes
 * Draw some small boxesin the corner of the buffer
 * based on some performance information
 */
static void performanceBoxes( void ) {
	if ( !mgaglx.boxes ) {
		return;
	}
	
	/* make sure we are drawing to the correct place */
	ENTER3D;
	
	/* draw a white box to show we are active, so if it is't seen
	it means that it is completely software based rendering  */
	ClearBox( 4, 4, 8, 8, 255, 255, 255 );
	
	/* draw a red box if we had to wait for drawing to complete 
	(software render or texture swap) */
	if ( mgaglx.c_drawWaits ) {
		ClearBox( 16, 4, 8, 8, 255, 0, 0 );
		mgaglx.c_drawWaits = 0;
	}
	
	/* draw a blue box if the register protection signal was hit */
	if ( mgaglx.c_signals ) {
		ClearBox( 28, 4, 8, 8, 0, 0, 255 );
		mgaglx.c_signals = 0;
	}
	
	/* draw a yellow box if textures were swapped */
	if ( mgaglx.c_textureSwaps ) {
		ClearBox( 40, 4, 8, 8, 255, 255, 0 );
		mgaglx.c_textureSwaps = 0;
	}
	
	/* draw a purple box if we flushed more than one buffer */
	if ( mgaglx.c_dmaFlsuh > 1 ) {
		ClearBox( 52, 4, 8, 8, 0, 255, 255 );
	}
	mgaglx.c_dmaFlsuh = 0;
	
	/* draw a green box if we had to wait for dma to complete (full utilization) 
	on the previous frame */
	if ( mgaglx.fullUtilization && !mgaglx.c_drawWaits ) {
		ClearBox( 64, 4, 8, 8, 0, 255, 0 );
		mgaglx.c_drawWaits = 0;
	}
}

/*
 * Copy the back buffer to the front buffer.  If there's no back buffer
 * this is a no-op.
 */
void mgaGLXSwapBuffers( XSMesaBuffer b ) {
#ifdef PROFILE
	GLdouble t0 = gl_time();
#endif
	mgaBufferPtr buf;

	/* make sure mesa gives us everything */
	/* 3.1 leaves unsent verts without this, but there is probably
	a better internal call to get them sent... */
	glFlush();

	/* report performance counters */
	mgaMsg(10, "swapBuffers: c_triangles:%i  c_lines:%i c_points:%i c_setupPointers:%i\n", 
		mgaglx.c_triangles, mgaglx.c_lines, mgaglx.c_points, mgaglx.c_setupPointers );
	mgaglx.c_triangles = 0;
	mgaglx.c_lines = 0;
	mgaglx.c_points = 0;
	mgaglx.c_setupPointers = 0;

	if ( !b->db_state ) {
		/* not double buffered, so do nothing */
	} else {
		ValidateGC(b->frontbuffer, b->cleargc);
		if ( b->backimage ) {
			if ( (buf = (mgaBufferPtr)b->backimage->devPriv) != NULL && buf->Drawable ) {
				/* diagnostic drawing tools */
				performanceBoxes();

				/* hardware accelerated back to front blit */	
				BackToFront( (DrawablePtr)b->frontbuffer,buf );

				/* make sure all dma is going to get executed */
				/* if everything has gone well, this will be the only flush each frame */
				mgaDmaFlush();
			} else {
				/* Use backimage's dimension (not buffer's) */
				(*b->cleargc->ops->PutImage)((DrawablePtr)b->frontbuffer,
							 b->cleargc,
							 b->frontbuffer->depth,
							 0, 0,
							 b->backimage->width, 
							 b->backimage->height,
							 0, ZPixmap, b->backimage->data);
			}
		} else {
			/* Copy pixmap to window on server */
			(*b->cleargc->ops->CopyArea)((DrawablePtr)b->backpixmap,
					       b->frontbuffer,
					       b->cleargc,
					       0, 0, b->width, b->height,
					       0, 0);
		}
    }
  
#ifdef PROFILE
  XSMesa->gl_ctx->SwapCount++;
  XSMesa->gl_ctx->SwapTime += gl_time() - t0;
#endif
}

/*
 * DoMakeCurrent
 * Changing either the context or the buffer will force this.
 */
static int DoMakeCurrent(XSMesaContext c, XSMesaBuffer b) {
	mgaContextPtr ctx;
	mgaBufferPtr	buf;
  
	mgaMsg( 10, "DoMakeCurrent( %p, %p )\n", c, b );

	/* we will need to program new destination registers before we draw */
	mgaglx.needEnter3D = 1;

	mgaCtx = NULL;
	mgaDB = NULL;
	
	if ( c ) {
		ctx = c->hw_ctx;
		if (!VALID_MGA_CONTEXT(ctx)) {
			return -1;
		}
	} else {
		ctx = NULL;
	}
	
	if ( b && b->buffer == XIMAGE ) {
		buf = (mgaBufferPtr) b->backimage->devPriv;
		if (buf->magic != mgaBufferMagic) {
			return -1;
		}
		if (!buf->Drawable) {
			buf = NULL;
		}
	} else {
		buf = NULL;
	}

	if ( !ctx && buf ) {
		return -1;
	}
	if (!ctx) {
		return 0;
	}
	
	ctx->DB = buf;

	/* enable clipping */
	MGA_SET_FIELD(ctx->regDWGCTL,DC_clipdis_MASK,DC_clipdis_disable); 

	/* valid conditions to render with */
	mgaCtx = ctx;
	mgaDB = ctx->DB;

	return 0;
}


/*
 * mgaGLXMakeCurrent
 * Note that GLX calls this every batch of commands,
 * so it will show up several times a frame even if
 * the client application isn't issuing makeCurrent calls.
 */ 
GLboolean mgaGLXMakeCurrent( XSMesaContext c ) {
	if (c == XSMesa) {
		return (GL_TRUE);
	}
	
	/* mesa 3.1 will sometimes leave unflushed vertexes */
	if ( XSMesa ) {
		glFlush();
	}

 	mgaMsg( 10, "mgaGLXMakeCurrent( %p )\n", c );
	XSMesa = c;                                                                 
	if (c) {
		DoMakeCurrent(c,c->xsm_buffer);
		gl_make_current(c->gl_ctx, 
		  c->xsm_buffer ? c->xsm_buffer->gl_buffer : NULL);
	} else {
	  	/* detach */
		DoMakeCurrent(NULL,NULL);
		gl_make_current(NULL, NULL);
	}
	return(GL_TRUE); 
}

/*
 * Bind buffer b to context c and make c the current rendering context.
 */
GLboolean mgaGLXBindBuffer( XSMesaContext c, XSMesaBuffer b ) {
	mgaMsg( 10, "mgaGLXBindBuffer( %p, %p )\n", c, b );
	DoMakeCurrent(c,b);
	return(XSMesaBindBuffer(c,b));
}


/*
 * Create a new XSMesaContext.
 * Input:  v - XSMesaVisual
 *         share_list - another XSMesaContext with which to share display
 *                      lists or NULL if no sharing is wanted.
 * Return:  an XSMesaContext or NULL if error.
 */
XSMesaContext mgaGLXCreateContext( XSMesaVisual v,  
                                   XSMesaContext share_list )
{
    XSMesaContext c;

    mgaMsg (0, "### Creating new xsmesa context...\n" );

    c = (XSMesaContext) calloc( 1, sizeof(struct xsmesa_context) );
    if (!c)
    {
        return(NULL);
    }

    c->gl_ctx = gl_create_context(v->gl_visual,    
				  share_list ? share_list->gl_ctx : NULL,
				  (void *) c, GL_TRUE /* direct rendering */);
    if (!c->gl_ctx)
    {
        free(c);
        return(NULL);
    }

    c->xsm_visual = v;
    c->xsm_buffer = NULL;       /* set later by XSMesaMakeCurrent */
    c->pixelformat = v->dithered_pf;    /* Dithering is enabled by default */
    c->hw_ctx = (void *)mgaCreateContext(c->gl_ctx);
    if (!c->hw_ctx) {
      mgaError("Cannot create hardware specific context.\n");
      return (NULL);
  }
    ((mgaContextPtr)(c->hw_ctx))->refcount++;
    c->gl_ctx->Driver.UpdateState = mga_setup_DD_pointers;

    mgaMsg( 1, "mgaGLXCreateContext succeeded: %p\n", c );

    return(c);
}

void mgaGLXDestroyContext( XSMesaContext c ) {
	mgaMsg( 1, "mgaGLXDestroyContext( %p )\n", c );
	
	/* make sure all drawing is completed */
	mgaWaitDrawingEngine();

	if (mgaDestroyContext((mgaContextPtr) c->hw_ctx) != 0) {
		mgaError("mgaGLXDestroyContext(): mgaDestroyContext() failed!\n");
	}
	XSMesaDestroyContext(c);
	if ( XSMesa == c ) {
		XSMesa = 0;
	}
}

/*
 * This function should only verify that the current hardware is supported.
 * It should do no setup. As we support various Matrox chipsets, perhaps it
 * should return an indicator of which chipset is present.
 */
GLboolean det_hwGfx() {  
    mgaMsg(1,"Detected 0x%x Chip ID\n", MGAchipset);
   /* is this the best way check for mga presence? */
   if(!MGA_IS_G200(MGAchipset) && !MGA_IS_G400(MGAchipset)){ 
      mgaError("MGAchipset not set, no mga hardware?\n");
      return GL_FALSE;
   }

   /* FIXME: implement support for other depths... */
   if(vga256InfoRec.depth != 15 &&
      vga256InfoRec.depth != 16 &&
      vga256InfoRec.depth != 24 ) {
      mgaError("Unsupported depth: %d, only 15,16, and 24 bpp are supported right now\n",
	       vga256InfoRec.depth);
      return GL_FALSE;
   }

   return GL_TRUE;
}


/*
 * mgaInitGLX
 * This is the initial entry point for the mga hardware driver,
 * called at module load time.
 */
GLboolean mgaInitGLX( void ) {
	mgaOpenLog(getenv("GLX_MGA_LOGFILE"));
	if (getenv("GLX_MGA_LOGLEVEL")) {
		mgaSetLogLevel(atoi(getenv("GLX_MGA_LOGLEVEL")));
	} else {
		mgaSetLogLevel(DBG_LEVEL_BASE);
	}

	mgaMsg(1,"virtual (x, y) (%d, %d)\n", vga256InfoRec.virtualX,
		 vga256InfoRec.virtualY);
	mgaMsg(1,"width: %d\n", vga256InfoRec.displayWidth);
	mgaMsg(1,"depth: %d\n", vga256InfoRec.depth);
	mgaMsg(1,"memBase: 0x%08x\n", vgaLinearBase);
	mgaMsg(1,"videoRam: 0x%08x\n", vga256InfoRec.videoRam);

	/* check to make sure that we are on an apropriate chip and not
	running in 8bpp mode */
	if ( !det_hwGfx() ) {
		return GL_FALSE;
	}

	/* start up our card memory manager */
	cardHeap = mmInit(vga256InfoRec.videoRam * 1024);
	if ( !cardHeap ) {
		mgaMsg( 1,"cardHeap creation failed, exiting!\n" );
		return GL_FALSE;	/* really shouldn't happen */
	}
	cardPhysical = (mgaUI32)xf86AccelInfoRec.ServerInfoRec->physBase;
	cardVirtual = (unsigned char *)vgaLinearBase;
	
	/* reserve memory used by the desktop screen and set up
	some hardware acceleration values needed to draw to it */
	mgaFrontBuffer = CreateFrontBuffer();
	if (!mgaFrontBuffer) {
		mgaError("Cannot create front buffer.\n");
		return GL_FALSE;	/* really shouldn't happen */
	}
	/* reserve memory X uses for pixmap acceleration */
	mmReserveMem( cardHeap, xf86AccelInfoRec.PixmapCacheMemoryStart,
	       xf86AccelInfoRec.PixmapCacheMemoryEnd -
	       xf86AccelInfoRec.PixmapCacheMemoryStart);

	/* reserve last 1KB if hardware cursor is active*/
	if (MGAdac.isHwCursor) {	
		mmReserveMem( cardHeap, (vga256InfoRec.videoRam-1)*1024,1024);
	}
	/* the remaining memory is available for back buffers, depth
	buffers, and textures */
	mmDumpMemInfo( cardHeap );

	/* init the dma system */  
	mgaLibInit();

	/* FIXME: what other GLXProcs pointers should we change? */
	GLXProcs.CreateContext = mgaGLXCreateContext;
	GLXProcs.DestroyContext = mgaGLXDestroyContext;
	GLXProcs.SwapBuffers = mgaGLXSwapBuffers;
	GLXProcs.CreateImage = mgaGLXCreateImage;
	GLXProcs.DestroyImage = mgaGLXDestroyImage;
	GLXProcs.CreateDepthBuffer = mgaGLXCreateDepthBuffer;
	GLXProcs.MakeCurrent = mgaGLXMakeCurrent;
	GLXProcs.BindBuffer = mgaGLXBindBuffer;

	/* for performance tuning, skip all of the triangle setup routines */  
	if (getenv("GLX_MGA_NULLPRIMS") ) {
		mgaMsg( 1, "enabling GLX_MGA_NULLPRIMS\n" );
		mgaglx.nullprims = 1;
	}
	if (getenv("GLX_MGA_BOXES") ) {
		mgaMsg( 1, "enabling GLX_MGA_BOXES\n" );
		mgaglx.boxes = 1;
	}
	if (getenv("GLX_MGA_NOFALLBACK") ) {
		mgaMsg( 1, "enabling GLX_MGA_NOFALLBACK\n" );
		mgaglx.noFallback = 1;
	}

	mgaError("mgaInitGLX completed\n");
	return GL_TRUE;
}


