/*
 * GLX Hardware Device Driver for Matrox Millenium G200
 * Copyright (C) 1999 Wittawat Yamwong
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * WITTAWAT YAMWONG, OR ANY OTHER CONTRIBUTORS BE LIABLE FOR ANY CLAIM, 
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE 
 * OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 *
 *    Wittawat Yamwong <Wittawat.Yamwong@stud.uni-hannover.de>
 */

/* $Id: mgalib.h,v 1.1 1999/07/31 07:07:55 crow Exp $ */

#ifndef MGALIB_INC
#define MGALIB_INC

#include "mgacommon.h"
#include "mgabuf.h"
#include "mgacontext.h"
#include "mm.h"
#include "g200_mac.h"
#include "mgalog.h"
#include "mgadma.h"

extern mgaBufferPtr mgaDB;
extern mgaBufferPtr mgaFrontBuffer;
extern mgaContextPtr mgaCtx;
extern int mgaIn3DContext;

typedef struct {
	mgaBufferPtr	bufferList;

	/* when set, the 3D register state will be programmed before the next primitive */
	mgaUI32		needEnter3D;

	/* dma stuff */
	mgaUI32		dmaDriver;	
	
	/* options */
	mgaUI32		nullprims;		/* skip all primitive generation */
	mgaUI32		boxes;			/* draw performance boxes */
	mgaUI32		noFallback;		/* don't fall back to software, do best-effort rendering */
	
	/* performance counters */	
	mgaUI32		c_setupPointers;
	mgaUI32		c_triangles;
	mgaUI32		c_points;
	mgaUI32		c_lines;
	mgaUI32		c_drawWaits;
	mgaUI32		c_textureSwaps;
	mgaUI32		c_signals;
	mgaUI32		c_dmaFlsuh;
	mgaUI32		c_overflows;
	
	mgaUI32		fullUtilization;
} mgaGlx_t;

extern mgaGlx_t	mgaglx;

void mgaLibInit();
void mgaSoftReset();

int mgaMakeCurrent(mgaContextPtr ctx, mgaBufferPtr buf);
void mgaLoadTexturePalette(mgaUI16 *pal, mgaUI8 start, mgaUI16 len);

#define MGAPACKCOLOR555(r,g,b,a) \
  ((((r) & 0xf8) << 7) | (((g) & 0xf8) << 2) | (((b) & 0xf8) >> 3) | \
    ((a) ? 0x8000 : 0))

#define MGAPACKCOLOR565(r,g,b) \
  ((((r) & 0xf8) << 8) | (((g) & 0xfc) << 3) | (((b) & 0xf8) >> 3))

#define MGAPACKCOLOR888(r,g,b) \
  (((r) << 16) | ((g) << 8) | (b))

#define MGAPACKCOLOR8888(r,g,b,a) \
  (((a) << 24) | ((r) << 16) | ((g) << 8) | (b))

#define MGAPACKCOLOR4444(r,g,b,a) \
  ((((a) & 0xf0) << 8) | (((r) & 0xf0) << 4) | ((g) & 0xf0) | ((b) >> 4))

static __inline__ mgaUI32 mgaPackColor(mgaUI8 r, mgaUI8 g, 
				   mgaUI8 b, mgaUI8 a)
{
  switch (mgaDB->Attrib & MGA_PF_MASK) {
  case MGA_PF_555:
    return MGAPACKCOLOR555(r,g,b,a);
  case MGA_PF_565:
    return MGAPACKCOLOR565(r,g,b);
  case MGA_PF_888:
    return MGAPACKCOLOR888(r,g,b);
  case MGA_PF_8888:
    return MGAPACKCOLOR8888(r,g,b,a);
  default:
    return 0;
  }
}

#endif
