/*
 * GLX Hardware Device Driver for Matrox Millenium G200
 * Copyright (C) 1999 Wittawat Yamwong
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * WITTAWAT YAMWONG, OR ANY OTHER CONTRIBUTORS BE LIABLE FOR ANY CLAIM, 
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE 
 * OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 *
 *    Wittawat Yamwong <Wittawat.Yamwong@stud.uni-hannover.de>
 */

/* $Id: mgalines.c,v 1.1 1999/07/31 07:07:55 crow Exp $ */

#include "mesaglx/vb.h"

#include "g200_mac.h"
#include "mm.h"
#include "mgalib.h"
#include "mgadma.h"
#include "mgalines.h"
#include <stdlib.h>

#define FLIP(x) (mgaDB->Height - (x) - 1)

/*
 * READ THIS:
 * You may want to soft reset the graphic chip at the first time, so that
 * line functions will work properly.
 * See mgaSoftReset() in mgalib.c
 */

/* TODO: fog */


static void simple_line_z( GLcontext *ctx, GLuint v1, GLuint v2, GLuint pv )
{
  struct vertex_buffer *VB = ctx->VB;
  mgaI32 dx,dy;
  mgaI32 x1,y1,x2,y2;
  mgaI32 dwgctl;
  mgaI32 zstart,zstep;
DMALOCALS;

	/* bump the performance counter */
 	mgaglx.c_lines++;

	/* setup 3D state if needed */
	ENTER3D;

  MGADMAGETPTR(75);
   
  x1 = (mgaI32) VB->VBWIN[v1][0];
  y1 = FLIP((mgaI32) VB->VBWIN[v1][1]);
  x2 = (mgaI32) VB->VBWIN[v2][0];
  y2 = FLIP((mgaI32) VB->VBWIN[v2][1]);

  dx = x2 - x1; if (dx < 0) dx = -dx;
  dy = y2 - y1; if (dy < 0) dy = -dy;
  if (dx == 0 && dy == 0)
    return;
  if (dy > dx)
    dx = dy;
  /* dx := max(dx,dy) */

  MGADMA_ALPHACTRL(mgaCtx->regALPHACTRL);
  if (ctx->Light.ShadeModel == GL_FLAT) {
    MGADMA_R(mgaF0915(VB->VBCOLOR[pv][0]),0,0);
    MGADMA_G(mgaF0915(VB->VBCOLOR[pv][1]),0,0);
    MGADMA_B(mgaF0915(VB->VBCOLOR[pv][2]),0,0);
    MGADMA_A(mgaF0915(VB->VBCOLOR[pv][3]),0,0);
  } else {
    mgaI32 start,step;
    start = VB->VBCOLOR[v1][0] << 15;
    step = (((VB->VBCOLOR[v2][0] << 15) - start) / dx) & 0xffffff;
    MGADMA_R(start,step,step);
    start = VB->VBCOLOR[v1][1] << 15;
    step = (((VB->VBCOLOR[v2][1] << 15) - start) / dx) & 0xffffff;
    MGADMA_G(start,step,step);
    start = VB->VBCOLOR[v1][2] << 15;
    step = (((VB->VBCOLOR[v2][2] << 15) - start) / dx) & 0xffffff;
    MGADMA_B(start,step,step);
    start = VB->VBCOLOR[v1][3] << 15;
    step = (((VB->VBCOLOR[v2][3] << 15) - start) / dx) & 0xffffff;
    MGADMA_A(start,step,step);
  }

  dwgctl = mgaCtx->regDWGCTL;
  MGA_SET_FIELD(dwgctl,
                DC_opcod_MASK & DC_bltmod_MASK & DC_atype_MASK & DC_zmode_MASK,
		DC_opcod_autoline_open | DC_bltmod_bfcol);
  
  zstart = (mgaI32) (VB->VBWIN[v1][2] * (1 << 15));
  zstep = ((mgaI32)(VB->VBWIN[v2][2] * (1 << 15)) - zstart) / dx;
  MGADMA_Z16(zstart,zstep,zstep);
  dwgctl |= mgaCtx->zmode;
  MGADMA_AUTO_LINE(x1,y1,x2,y2);
  MGADMA_DWGCTL_EXEC(dwgctl);
  DMAADVANCE();
}


static void simple_line_noz( GLcontext *ctx, GLuint v1, GLuint v2, GLuint pv )
{
  struct vertex_buffer *VB = ctx->VB;
  mgaI32 dx,dy;
  mgaI32 x1,y1,x2,y2;
  mgaI32 dwgctl;
DMALOCALS;
   
	/* bump the performance counter */
 	mgaglx.c_lines++;

	/* setup 3D state if needed */
	ENTER3D;
   
  MGADMAGETPTR(75);

  x1 = (mgaI32) VB->VBWIN[v1][0];
  y1 = FLIP((mgaI32) VB->VBWIN[v1][1]);
  x2 = (mgaI32) VB->VBWIN[v2][0];
  y2 = FLIP((mgaI32) VB->VBWIN[v2][1]);
  dx = x2 - x1; if (dx < 0) dx = -dx;
  dy = y2 - y1; if (dy < 0) dy = -dy;
  if (dx == 0 && dy == 0)
    return;
  if (dy > dx)
    dx = dy;
  /* dx := max(dx,dy) */

  MGADMA_ALPHACTRL(mgaCtx->regALPHACTRL);
  if (ctx->Light.ShadeModel == GL_FLAT) {
    MGADMA_R(mgaF0915(VB->VBCOLOR[pv][0]),0,0);
    MGADMA_G(mgaF0915(VB->VBCOLOR[pv][1]),0,0);
    MGADMA_B(mgaF0915(VB->VBCOLOR[pv][2]),0,0);
    MGADMA_A(mgaF0915(VB->VBCOLOR[pv][3]),0,0);
  } else {
    mgaI32 start,step;
    start = VB->VBCOLOR[v1][0] << 15;
    step = (((VB->VBCOLOR[v2][0] << 15) - start) / dx) & 0xffffff;
    MGADMA_R(start,step,step);
    start = VB->VBCOLOR[v1][1] << 15;
    step = (((VB->VBCOLOR[v2][1] << 15) - start) / dx) & 0xffffff;
    MGADMA_G(start,step,step);
    start = VB->VBCOLOR[v1][2] << 15;
    step = (((VB->VBCOLOR[v2][2] << 15) - start) / dx) & 0xffffff;
    MGADMA_B(start,step,step);
    start = VB->VBCOLOR[v1][3] << 15;
    step = (((VB->VBCOLOR[v2][3] << 15) - start) / dx) & 0xffffff;
    MGADMA_A(start,step,step);
  }

  dwgctl = mgaCtx->regDWGCTL;
  MGA_SET_FIELD(dwgctl,
                DC_opcod_MASK & DC_bltmod_MASK & DC_atype_MASK & DC_zmode_MASK,
		DC_opcod_autoline_open | DC_bltmod_bfcol);
  dwgctl |= DC_atype_i | DC_zmode_nozcmp;
  MGADMA_AUTO_LINE(x1,y1,x2,y2);
  MGADMA_DWGCTL_EXEC(dwgctl);
  DMAADVANCE();
}


line_func mga_choose_line_function( GLcontext *ctx ) {
	if ( !mgaglx.noFallback ) {
		if ( ctx->Texture.Enabled ) {
			mgaMsg( 10, "software fallback for textured lines\n" );
			return NULL;
		}
		if ( ctx->Stencil.Enabled ) {
			mgaMsg( 10, "software fallback for stenciled lines\n" );
			return NULL;
		}
		if ( ctx->Line.StippleFlag ) {
			mgaMsg( 10, "software fallback for stippled lines\n" );
			return NULL;
		}
		if ( ctx->Line.SmoothFlag ) {
			mgaMsg( 10, "software fallback for smooth lines\n" );
			return NULL;
		}
		if ( ctx->Line.Width != 1.0F ) {
			mgaMsg( 10, "software fallback for Line.Width != 1.0F\n" );
			return NULL;
		}

		if (
		#ifdef MESA31 
		  ctx->RasterMask & MULTI_DRAW_BIT
		#else      
		  ctx->RasterMask & FRONT_AND_BACK_BIT
		#endif      
		) {
			mgaMsg( 10, "software fallback for MULTI_DRAW_BIT lines\n" );
			return NULL;
		}
	}
	if ( mgaCtx->NoZ ) {
		return simple_line_noz;
	}
	return simple_line_z;
}
