
/*
 * GLX Hardware Device Driver for Matrox Millenium G200
 * Copyright (C) 1999 Wittawat Yamwong
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * WITTAWAT YAMWONG, OR ANY OTHER CONTRIBUTORS BE LIABLE FOR ANY CLAIM, 
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE 
 * OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 *
 *    7/10/99 : created by John Carmack <johnc@idsoftware.com>
 */

#include "mesaglx/vb.h"

#include "g200_mac.h"
#include "mm.h"
#include "mgalib.h"
#include "mgadma.h"
#include "mgalines.h"
#include <stdlib.h>

#define FLIP(x) (mgaDB->Height - (x) - 1)

/* Drawing the points with the hardware like this isn't very fast, but it
is necessary to allow everything to continue asyncronously.  If anyone has
a real need for high speed point rendering, this can be made quite a bit
more efficient or special cased. */


static void points_depth( GLcontext *ctx, GLuint first, GLuint last ) {
	DMALOCALS;
	mgaUI32	cmd;
	struct vertex_buffer *VB;
	mgaVertexPtr		mgaVB;
	mgaI32	x, y, offset, size, w, h;
	mgaUI32	i;
	GLfloat Zstart;
	
	/* bump the performance counter */
	mgaglx.c_points += last - first + 1;

	/* setup 3D state if needed */
	ENTER3D;
	
	/* get the depth compare mode */
     	cmd = mgaCtx->zmode;
     	
	cmd |= DC_opcod_trap | DC_arzero_enable | DC_sgnzero_enable | 
	  DC_shftzero_enable | (0xC << DC_bop_SHIFT) | 
	  DC_atype_rstr | DC_solid_enable;

	VB = ctx->VB;
	mgaVB = mgaCtx->VB;

	size = ctx->Point.Size;
	if ( size < 1 ) {
		size = 1;
	}
	offset = size >> 1;

	/* set up for constant depth and color */	
	MGADMAGETPTR( 32 );

	DMAOUTREG(MGAREG_DR2,0);
	DMAOUTREG(MGAREG_DR3,0);
	
	DMAOUTREG(MGAREG_DR6,0);
	DMAOUTREG(MGAREG_DR7,0);
	
	DMAOUTREG(MGAREG_DR10,0);
	DMAOUTREG(MGAREG_DR11,0);
	
	DMAOUTREG(MGAREG_DR14,0);
	DMAOUTREG(MGAREG_DR15,0);
	
	DMAOUTREG(MGAREG_ALPHAXINC,0);
	DMAOUTREG(MGAREG_ALPHAYINC,0);
	
	DMAADVANCE();

	for ( i = first ; i <= last ; i++ ) {
		x = (mgaI32) VB->VBWIN[i][0] - offset;
		if ( x < 0 ) {
			w = size + x;
			if ( w <= 0 ) {
				continue;
			}
			x = 0;
		} else {
			w = size;
		}
		
		y = FLIP((mgaI32) VB->VBWIN[i][1]) - offset;
		if ( y < 0 ) {
			h = size + y;
			if ( y <= 0 ) {
				continue;
			}
			y = 0;
		} else {
			h = size;
		}
		MGADMAGETPTR( 32 );

		Zstart = mgaVB[i].z;
#if DEPTH_BITS==16
		DMAOUTREG(MGAREG_DR0,Zstart);
#else
#error 32 bits depth buffer not yet supported
#endif
		DMAOUTREG(MGAREG_DR4,VB->VBCOLOR[i][0] << 15);
		DMAOUTREG(MGAREG_DR8,VB->VBCOLOR[i][1] << 15);
		DMAOUTREG(MGAREG_DR12,VB->VBCOLOR[i][2] << 15);
		DMAOUTREG(MGAREG_ALPHASTART,VB->VBCOLOR[i][3] << 15);

		MGADMA_YDSTLEN(y,h);
		MGADMA_FXBNDRY(x,x+w);
		MGADMA_DWGCTL_EXEC(cmd);

		DMAADVANCE();
	}
}


points_func mga_choose_points_function( GLcontext *ctx ) {
	if ( !mgaglx.noFallback ) {
		if ( ctx->Point.SmoothFlag ) {
			mgaMsg( 10, "software fallback for smooth points\n" );
			return NULL;
		}
		if ( ctx->Texture.Enabled ) {
			mgaMsg( 10, "software fallback for textured points\n" );
			return NULL;
		}
		if ( ctx->Stencil.Enabled ) {
			mgaMsg( 10, "software fallback for stenciled points\n" );
			return NULL;
		}
		/* FIXME: do we need to test for fog? */

		if (
		#ifdef MESA31 
		  ctx->RasterMask & MULTI_DRAW_BIT
		#else      
		  ctx->RasterMask & FRONT_AND_BACK_BIT
		#endif      
		) {
			mgaMsg( 10, "software fallback for MULTI_DRAW_BIT points\n" );
			return NULL;
		}
	}

	/* it isn't worth special casing for a non-depth setup point */	
	return points_depth;
}
