/*
 * GLX Hardware Device Driver for Matrox Millenium G200
 * Copyright (C) 1999 Wittawat Yamwong
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * WITTAWAT YAMWONG, OR ANY OTHER CONTRIBUTORS BE LIABLE FOR ANY CLAIM, 
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE 
 * OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 *
 *    Wittawat Yamwong <Wittawat.Yamwong@stud.uni-hannover.de>
 */

/* $Id: mgatex.h,v 1.1 1999/07/31 07:07:55 crow Exp $ */

#ifndef MGATEX_INC
#define MGATEX_INC

#include "mesaglx/types.h"

#include "mgacommon.h"
#include "mgabuf.h"

#define MGA_TEXTURE_OBJECT_MAGIC 0x9a0b3fd2
#define VALID_MGA_TEXTURE_OBJECT(tobj) \
  ((tobj) && (tobj->magic == MGA_TEXTURE_OBJECT_MAGIC))

/* for TexReg */
#define MGA_TEXREG_SIZE 16+4	/* includes three nops at the end */
#define MGA_TEXREG_0            0x968f8c97
#define MGA_TEXREG_TEXBORDERCOL 1
#define MGA_TEXREG_TEXCTL       2
#define MGA_TEXREG_TEXCTL2      3
#define MGA_TEXREG_TEXFILTER    4
#define MGA_TEXREG_5            0x8e8d8a8b
#define MGA_TEXREG_TEXHEIGHT    6
#define MGA_TEXREG_TEXWIDTH     7
#define MGA_TEXREG_TEXTRANS     8
#define MGA_TEXREG_TEXTRANSHIGH 9
#define MGA_TEXREG_10           0xabaaa989
#define MGA_TEXREG_TEXORG       11
#define MGA_TEXREG_TEXORG1      12
#define MGA_TEXREG_TEXORG2      13
#define MGA_TEXREG_TEXORG3      14
#define MGA_TEXREG_15           0x151515ac
#define MGA_TEXREG_TEXORG4      16

#define MGA_TEX_MAXLEVELS 5

/* number of fractional bits for s/w,t/w,q/w */
#define MGA_TEX_SBIT 20
#define MGA_TEX_TBIT 20
#define MGA_TEX_QBIT 20

struct mga_texture_object_t {
  mgaUI32 magic;
  struct mga_texture_object_t *next;

  mgaUI32 TexReg[MGA_TEXREG_SIZE];
  mgaBufferPtr Tex[MGA_TEX_MAXLEVELS];
  struct gl_texture_object *globj;
  int Complete;
  int swapped_out;
  unsigned long age;   
  mgaUI16 Palette[256];
  int UsePalette;
  int NullTexture;
};		
#define MGA_NO_PALETTE      0
#define MGA_USE_PALETTE     1
#define MGA_UPDATE_PALETTE  2

typedef struct mga_texture_object_t *mgaTextureObjectPtr;

mgaTextureObjectPtr mgaCreateTexObj(mgaContextPtr ctx,
				    struct gl_texture_object *tObj);
int mgaDestroyTexObj(mgaContextPtr ctx, mgaTextureObjectPtr t);
void mgaUpdateTextureState(GLcontext *ctx, mgaTextureObjectPtr t);
void mgaSetTexWrapping(mgaTextureObjectPtr tex, GLenum s, GLenum t);
void mgaSetTexFilter(mgaTextureObjectPtr t, GLenum minf, GLenum magf);
void mgaSetTexBorderColor(mgaTextureObjectPtr t, GLint color[4]);


/* Driver functions */

void mgaTexEnv( GLcontext *ctx, GLenum pname, const GLfloat *param );

void mgaTexImage( GLcontext *ctx, GLenum target,
		  struct gl_texture_object *tObj, GLint level,
		  GLint internalFormat,
		  const struct gl_texture_image *image );

void mgaTexSubImage( GLcontext *ctx, GLenum target,
		  struct gl_texture_object *tObj, GLint level,
		  GLint xoffset, GLint yoffset,
		  GLsizei width, GLsizei height,
		  GLint internalFormat,
		  const struct gl_texture_image *image );

void mgaTexParameter( GLcontext *ctx, GLenum target,
		      struct gl_texture_object *tObj,
		      GLenum pname, const GLfloat *params );

void mgaBindTexture( GLcontext *ctx, GLenum target,
		     struct gl_texture_object *tObj );

void mgaDeleteTexture( GLcontext *ctx, struct gl_texture_object *tObj );

void mgaUpdateTexturePalette( GLcontext *ctx, struct gl_texture_object *tObj );
void mgaUseGlobalTexturePalette( GLcontext *ctx, GLboolean state );

#endif
