/* $Id: mgatris.c,v 1.1 1999/07/31 07:07:55 crow Exp $ */
/*
 * GLX Hardware Device Driver for Matrox Millenium G200
 * Copyright (C) 1999 Wittawat Yamwong
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * WITTAWAT YAMWONG, OR ANY OTHER CONTRIBUTORS BE LIABLE FOR ANY CLAIM, 
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE 
 * OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 *
 *    Wittawat Yamwong <Wittawat.Yamwong@stud.uni-hannover.de>
 */

#include "mesaglx/vb.h"

#include "g200_mac.h"
#include "mm.h"
#include "mgalib.h"
#include "mgatris.h"
#include "mgavb.h"
#include "mgalog.h"

#ifdef USE_X86_ASM
#define RINT(x) my_rint(x)
#define CEIL(x) my_ceil(x)
#define FLOOR(x) my_floor(x)
#else
#define RINT(x) ((mgaI32)(x))
#define CEIL(x) ((mgaI32)ceil(x))
#define FLOOR(x) ((mgaI32)floor(x))
#endif

static __inline__ long my_rint(const float x)
{
  register float arg = x;
  long result;
  __asm__ ("fistl %0" : "=m" (result) : "t" (arg));
  return result;
}

static __inline__ long my_ceil(const float x)
{
  register float arg = x;
  volatile long value;
  volatile short cw, cwtmp;

  __asm__ volatile ("fnstcw %0" : "=m" (cw) : );
  cwtmp = (cw & 0xf3ff) | 0x0800; /* rounding up */
  __asm__ volatile ("fldcw %1\n"
                    "fistl %0\n"
                    "fldcw %2"
                    : "=m" (value) 
                    : "m" (cwtmp), "m" (cw), "t" (arg));
  return value;
}

static __inline__ long my_floor(const float x)
{
  register float arg = x;
  volatile long value;
  volatile short cw, cwtmp;

  __asm__ volatile ("fnstcw %0" : "=m" (cw) : );
  cwtmp = (cw & 0xf3ff) | 0x0400;
  __asm__ volatile ("fldcw %1\n"
                    "fistl %0\n"
                    "fldcw %2"
                    : "=m" (value) 
                    : "m" (cwtmp), "m" (cw), "t" (arg));
  return value;
}

static void triangle_flat( GLcontext *ctx, GLuint v0,
                          GLuint v1, GLuint v2, GLuint pv )
{
#define INTERP_Z 1
#include "mga_tritemp.h"
}

static void triangle_smooth( GLcontext *ctx, GLuint v0,
                          GLuint v1, GLuint v2, GLuint pv )
{
#define INTERP_Z 1
#define INTERP_RGB 1
#define INTERP_ALPHA 1
#include "mga_tritemp.h"
}

static void triangle_flat_texture( GLcontext *ctx, GLuint v0,
                          GLuint v1, GLuint v2, GLuint pv )
{
#define INTERP_Z 1
#define INTERP_STUV 1
#define TEXTURE 1
#include "mga_tritemp.h"
}

static void triangle_smooth_texture( GLcontext *ctx, GLuint v0,
                          GLuint v1, GLuint v2, GLuint pv )
{
#define INTERP_Z 1
#define INTERP_RGB 1
#define INTERP_ALPHA 1
#define INTERP_STUV 1
#define TEXTURE 1
#include "mga_tritemp.h"
}

static void triangle_flat_texture_spec( GLcontext *ctx, GLuint v0,
                          GLuint v1, GLuint v2, GLuint pv )
{
#define INTERP_Z 1
#define INTERP_STUV 1
#define TEXTURE 1
#define FLAT_SPEC 1
#include "mga_tritemp.h"
}

static void triangle_smooth_texture_spec( GLcontext *ctx, GLuint v0,
                          GLuint v1, GLuint v2, GLuint pv )
{
#define INTERP_Z 1
#define INTERP_RGB 1
#define INTERP_ALPHA 1
#define INTERP_STUV 1
#define TEXTURE 1
#define INTERP_SPEC 1
#include "mga_tritemp.h"
}

#define VCOLOR(i)			\
  v[i].r = VB->VBCOLOR[i][0] << 15;	\
  v[i].g = VB->VBCOLOR[i][1] << 15;	\
  v[i].b = VB->VBCOLOR[i][2] << 15;	\
  v[i].a = VB->VBCOLOR[i][3] << 15

#define VSPEC(i)			\
  v[i].sr = VB->Specular[i][0] << 15;	\
  v[i].sg = VB->Specular[i][1] << 15;	\
  v[i].sb = VB->Specular[i][2] << 15

#define MAKETWOSIDE(fname,col,spec)				\
static void fname ## _twoside( GLcontext *ctx, GLuint v0,	\
                          GLuint v1, GLuint v2, GLuint pv )	\
{								\
  struct vertex_buffer *VB = ctx->VB;				\
  mgaVertexPtr v = mgaCtx->VB;					\
  if (col) {							\
    VCOLOR(v0);							\
    VCOLOR(v1);							\
    VCOLOR(v2);							\
  }								\
  if (spec) {							\
    VSPEC(v0);							\
    VSPEC(v1);							\
    VSPEC(v2);							\
  }								\
  fname(ctx,v0,v1,v2,pv);					\
}

MAKETWOSIDE(triangle_smooth,1,0)
MAKETWOSIDE(triangle_smooth_texture,1,0)
MAKETWOSIDE(triangle_smooth_texture_spec,1,1)

static void triangle_invalid_spec( GLcontext *ctx, GLuint v0,
                          GLuint v1, GLuint v2, GLuint pv )
{
  ErrorF("triangle_invalid_spec: separated specular color w/o texture??\n");
}

static triangle_func triangle_functions[] = {
  triangle_flat,
  triangle_smooth,
  triangle_flat_texture,
  triangle_smooth_texture,

  triangle_invalid_spec, /* flat_spec */
  triangle_invalid_spec, /* smooth_spec */
  triangle_flat_texture_spec,
  triangle_smooth_texture_spec,
  
  triangle_flat,
  triangle_smooth_twoside,
  triangle_flat_texture,
  triangle_smooth_texture_twoside,
  
  triangle_invalid_spec, /* flat_spec */
  triangle_invalid_spec, /* smooth_spec */
  triangle_flat_texture_spec,
  triangle_smooth_texture_spec_twoside
};

#define SMOOTH_BIT	1
#define TEXTURE_BIT	2
#define SPEC_BIT	4
#define TWOSIDE_BIT	8

triangle_func mga_choose_triangle_function(GLcontext *ctx) {
  int funcindex = 0;

	if ( !mgaglx.noFallback ) {
		if (ctx->Polygon.SmoothFlag ||
		  ctx->Polygon.StippleFlag ||
		  ctx->Stencil.Enabled ||
		#ifdef MESA31 
		   0 /* (ctx->RasterMask & MULTI_DRAW_BIT)  !!! FIXME !!! */ 
		#else      
		  (ctx->RasterMask & FRONT_AND_BACK_BIT)
		#endif      
		  )
		{
			#ifdef MESA31
			mgaMsg(5, " unsupported tri mode %d %d %d %d\n",
			   ctx->Polygon.SmoothFlag,
			   ctx->Polygon.StippleFlag,
			   (ctx->RasterMask & MULTI_DRAW_BIT),
			   ctx->Stencil.Enabled
			   );
			#else
			mgaMsg(5, " unsupported tri mode %d %d %d\n",
			   ctx->Polygon.SmoothFlag,
			   ctx->Polygon.StippleFlag,
			   ctx->Stencil.Enabled
			   );
			 
			#endif
			return NULL;
		}
  	}
  	
  if (ctx->Light.ShadeModel == GL_SMOOTH)
    funcindex |= SMOOTH_BIT;

  if (ctx->Texture.Enabled && !mgaCtx->CurrentTexObj->NullTexture)
    funcindex |= TEXTURE_BIT;

  if (ctx->Light.Model.ColorControl == GL_SEPARATE_SPECULAR_COLOR)
    funcindex |= SPEC_BIT;
    
  if (ctx->Light.Model.TwoSide)
    funcindex |= TWOSIDE_BIT;
  
  return triangle_functions[funcindex];
}
