/* $Id: mgavb.c,v 1.1 1999/07/31 07:07:55 crow Exp $ */
/*
 * GLX Hardware Device Driver for Matrox Millenium G200
 * Copyright (C) 1999 Wittawat Yamwong
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
 * WITTAWAT YAMWONG, OR ANY OTHER CONTRIBUTORS BE LIABLE FOR ANY CLAIM, 
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR 
 * OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE 
 * OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 *
 *    Wittawat Yamwong <Wittawat.Yamwong@stud.uni-hannover.de>
 */
#include "mgacontext.h"
#include "mgavb.h"
#include "mgalog.h"

#include "os.h"

extern mgaContextPtr mgaCtx;
extern mgaBufferPtr mgaDB;

#undef TEXTURE_BIT
#undef SPEC_BIT
#undef FOG_BIT
#undef SMOOTH_BIT

#define TEXTURE_BIT	1
#define SPEC_BIT	2
#define FOG_BIT		4
#define SMOOTH_BIT	8

/* FIXME: ldexp() from math.h seems to be slower than this macro -Wittawat */
#define LDEXP(x,y) ((x) * (1 << (y)))

#ifdef MESA31

#define TEX {						\
  int tex_size = VB->TexCoordPtr[0]->size;		\
  GLfloat invW = VB->VBWIN[i][3] * mgaCtx->InvWScale ;	\
  GLfloat *tc = &(VB->TexCoordPtr[0]->data[i][0]);	\
  v->s = LDEXP(tc[0] * invW,MGA_TEX_SBIT);		\
  if (tex_size>1)					\
    v->t = LDEXP(tc[1] * invW,MGA_TEX_TBIT);		\
  else v->t = 0;					\
  if (tex_size>3)					\
    v->v = LDEXP(tc[3] * invW,MGA_TEX_QBIT);		\
  else v->v = LDEXP(invW, MGA_TEX_QBIT);		\
}

#define SPC {					\
  GLubyte *spec = &(VB->Spec[0][i][0]);		\
  v->sr = spec[0] << 15;			\
  v->sg = spec[1] << 15;			\
  v->sb = spec[2] << 15;			\
}

#define FOG \
  v->fog = LDEXP((*mgaCtx->Fog.Factor)(mgaCtx,VB->EyePtr->data[i][2])*255.0F,15)

#define COL {					\
  GLubyte *col = &(VB->Color[0]->data[i][0]);	\
  v->r = col[0] << 15;				\
  v->g = col[1] << 15;				\
  v->b = col[2] << 15;				\
  v->a = col[3] << 15;				\
}

#else /* not MESA31 */

#define TEX {						\
  GLfloat invW = mgaCtx->InvWScale / VB->Clip[i][3];	\
  GLfloat *tc = &(VB->MultiTexCoord[0][i][0]);		\
  v->s = LDEXP(tc[0] * invW, MGA_TEX_SBIT);		\
  v->t = LDEXP(tc[1] * invW, MGA_TEX_TBIT);		\
  v->v = LDEXP(tc[3] * invW, MGA_TEX_QBIT);		\
}

#define SPC {				\
  GLubyte *spec = &(VB->Fspec[i][0]);	\
  v->sr = spec[0] << 15;		\
  v->sg = spec[1] << 15;		\
  v->sb = spec[2] << 15;		\
}

#define FOG \
  v->fog = LDEXP((*mgaCtx->Fog.Factor)(mgaCtx,VB->Eye[i][2])*255.0F,15)

#define COL {				\
  GLubyte *col = &(VB->Fcolor[i][0]);	\
  v->r = col[0] << 15;			\
  v->g = col[1] << 15;			\
  v->b = col[2] << 15;			\
  v->a = col[3] << 15;			\
}

#endif /* MESA31 */


#define NOP

#ifdef MESA31
#define MESA_PARAM  struct vertex_buffer *VB
#define MESA_VB
#else
#define MESA_PARAM GLcontext *ctx
#define MESA_VB struct vertex_buffer *VB = ctx->VB;
#endif


#ifdef USE_X86_ASM

static volatile short cw, cwtmp;

static __inline__ void ceil_start(void)
{
  __asm__ volatile ("fnstcw %0" : "=m" (cw) : );
  cwtmp = (cw & 0xf3ff) | 0x0800; /* rounding up */
  __asm__ volatile ("fldcw %0" : "=m" (cwtmp) : );
}
 
 static __inline__ long my_ceil(const float x) {
   register float arg = x;
  volatile long value;
 __asm__ volatile ("fistl %0\n"
                    : "=m" (value) 
                    : "t" (arg));
  return value;
}

 static __inline__ void ceil_end(void) {
 __asm__ volatile ("fldcw %0"
                    :
                    : "m" (cw) );
}
#define CEIL_START ceil_start()
#define CEIL_END ceil_end()
#define CEIL(x) my_ceil(x)
#else
#define CEIL(x) ((mgaI32)ceil(x))
#define CEIL_START
#define CEIL_END
#endif


#define SETUPFUNC(name,col,tex,spec,fog)		\
static void name(MESA_PARAM, GLuint start, GLuint end)	\
{							\
  MESA_VB						\
  mgaVertexPtr v;					\
  int i;						\
  float mgaDBHeightFloat;							\
 \
  CHECK_CONTEXT( return; );				\
  mgaDBHeightFloat = mgaDB->Height - 0.5f;	\
  CEIL_START;							\
  v = &(mgaCtx->VB[start]);				\
  for (i=start; i < end; i++) {				\
    if (VB->ClipMask[i] == 0) {				\
      GLfloat *win = &(VB->VBWIN[i][0]);		\
      v->idx = i;					\
      v->x = win[0] - 0.5F;				\
      v->y = mgaDBHeightFloat - win[1];		\
      v->ceilY = CEIL(v->y);			\
      v->z = LDEXP(win[2],15);				\
      col;						\
      tex;						\
      spec;						\
      fog;						\
    }							\
    v++;						\
  }							\
  CEIL_END;					\
}

SETUPFUNC(rs,		NOP,NOP,NOP,NOP)
SETUPFUNC(rs_t,		NOP,TEX,NOP,NOP)
SETUPFUNC(rs_f,		NOP,NOP,NOP,FOG)
SETUPFUNC(rs_ft,	NOP,TEX,NOP,FOG)
SETUPFUNC(rs_g,		COL,NOP,NOP,NOP)
SETUPFUNC(rs_gt,	COL,TEX,NOP,NOP)
SETUPFUNC(rs_gst,	COL,TEX,SPC,NOP)
SETUPFUNC(rs_gf,	COL,NOP,NOP,FOG)
SETUPFUNC(rs_gft,	COL,TEX,NOP,FOG)
SETUPFUNC(rs_gfst,	COL,TEX,SPC,FOG)

void rs_invalid(MESA_PARAM, GLuint start, GLuint end)
{
  mgaError("mgaRasterSetup(): invalid combination, "
	   "specular color requested w/o texture!\n");
}

/* 
   g = smooth shading
   t = texture
   s = specular color
   f = fog factor
*/
static mgaRasterSetupFunc setup_func[] = {
  rs,
  rs_t,
  rs_invalid,
  rs_t,
  rs_f,
  rs_ft,
  rs_invalid,
  rs_ft,
  rs_g,
  rs_gt,
  rs_invalid,
  rs_gst,
  rs_gf,
  rs_gft,
  rs_invalid,
  rs_gfst
};

mgaRasterSetupFunc mgaChooseRasterSetupFunc(GLcontext *ctx)
{
  int funcindex = 0;

  if (ctx->Texture.Enabled)
    funcindex |= TEXTURE_BIT;

  if (ctx->Light.Model.ColorControl == GL_SEPARATE_SPECULAR_COLOR)
    funcindex |= SPEC_BIT;

  if (ctx->Fog.Enabled)
    funcindex |= FOG_BIT;
    
  if (ctx->Light.ShadeModel == GL_SMOOTH && !ctx->Light.Model.TwoSide)
      funcindex |= SMOOTH_BIT;

  return setup_func[funcindex];
}
#undef MESA_PARAM
#undef MESA_VB
